"""
    Diagnostic streamfunction classes
    =================================

    Classes defining streamfunction fields diagnostics.

    Description of the classes
    --------------------------

    * :class:`AtmosphericStreamfunctionDiagnostic`: General base class for atmospheric streamfunction fields diagnostic.
    * :class:`LowerLayerAtmosphericStreamfunctionDiagnostic`: Diagnostic giving the lower layer atmospheric streamfunction fields :math:`\\psi^3_{\\rm a}`.
    * :class:`UpperLayerAtmosphericStreamfunctionDiagnostic`: Diagnostic giving the upper layer atmospheric streamfunction fields :math:`\\psi^1_{\\rm a}`.
    * :class:`MiddleAtmosphericStreamfunctionDiagnostic`: Diagnostic giving the middle atmospheric streamfunction fields :math:`\\psi_{\\rm a}`.
    * :class:`OceanicLayerStreamfunctionDiagnostic`: Diagnostic giving the oceanic layer streamfunction fields :math:`\\psi_{\\rm o}`.

"""

import warnings

import numpy as np
from scipy.integrate import dblquad
import matplotlib.pyplot as plt
from qgs.diagnostics.util import create_grid_basis

from qgs.diagnostics.base import FieldDiagnostic

# TODO: - convert the matmul and swapaxes into tensordot (cleaner)


class AtmosphericStreamfunctionDiagnostic(FieldDiagnostic):
    """General base class for atmospheric streamfunction fields diagnostic.
    Provide a spatial gridded representation of the fields.
    This is an `abstract base class`_, it must be subclassed to create new diagnostics!

    .. _abstract base class: https://docs.python.org/3/glossary.html#term-abstract-base-class

    Parameters
    ----------

    model_params: QgParams
        An instance of the model parameters.
    delta_x: float, optional
        Spatial step in the zonal direction `x` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    delta_y: float, optional
        Spatial step in the meridional direction `y` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    dimensional: bool
        Indicate if the output diagnostic must be dimensionalized or not.

    Attributes
    ----------

    dimensional: bool
        Indicate if the output diagnostic must be dimensionalized or not.
    """

    def __init__(self, model_params, delta_x=None, delta_y=None, dimensional=True):

        FieldDiagnostic.__init__(self, model_params, dimensional)
        self._configure(delta_x=delta_x, delta_y=delta_y)

        self._default_plot_kwargs['cmap'] = plt.get_cmap('jet')

    def _compute_grid(self, delta_x=None, delta_y=None):

        if delta_x is None:
            ams = self._model_params.ablocks
            if ams is None:
                warnings.warn("AtmosphericStreamfunctionDiagnostic: Unable to configure the grid automatically. Atmospheric wavenumbers information not " +
                              "present in the model's parameters ! Please call the compute_grid method with the delta_x and delta_y parameters.")
                return 1
            xwn = [ams[i][0] for i in range(len(ams))]
            mxwn = max(xwn)
            n_point_x = 4 * mxwn + 2
        else:
            n_point_x = int(np.ceil((2 * np.pi / self._model_params.scale_params.n) / delta_x) + 1)

        if delta_y is None:
            ams = self._model_params.ablocks
            if ams is None:
                warnings.warn("AtmosphericStreamfunctionDiagnostic: Unable to configure the grid automatically. Atmospheric wavenumbers information not " +
                              "present in the model's parameters ! Please call the compute_grid method with the delta_x and delta_y parameters.")
                return 1
            ywn = [ams[i][1] for i in range(len(ams))]
            mywn = max(ywn)
            n_point_y = 4 * mywn + 2
        else:
            n_point_y = int(np.ceil(np.pi / delta_y) + 1)

        x = np.linspace(0., 2 * np.pi / self._model_params.scale_params.n, n_point_x)
        y = np.linspace(0., np.pi, n_point_y)
        self._X, self._Y = np.meshgrid(x, y)

    def _configure(self, delta_x=None, delta_y=None):

        self._compute_grid(delta_x, delta_y)

        self._grid_basis = create_grid_basis(self._model_params.atmospheric_basis, self._X, self._Y, self._subs)
        if self._orography:
            if self._model_params.ground_params.orographic_basis == "atmospheric":
                self._oro_basis = self._grid_basis
            else:
                self._oro_basis = create_grid_basis(self._model_params.ground_basis, self._X, self._Y, self._subs)
        else:
            self._oro_basis = None


class LowerLayerAtmosphericStreamfunctionDiagnostic(AtmosphericStreamfunctionDiagnostic):
    """Diagnostic giving the lower layer atmospheric streamfunction fields :math:`\\psi^3_{\\rm a}`.
    Computed as :math:`\\psi^3_{\\rm a} = \\psi_{\\rm a} - \\theta_{\\rm a}` where :math:`\\psi_{\\rm a}` and :math:`\\theta_{\\rm a}` are respectively the barotropic and baroclinic streamfunctions.
    See also the :ref:`files/model/atmosphere:Atmospheric component` and :ref:`files/model/oro_model:Mid-layer equations and the thermal wind relation` sections.

    Parameters
    ----------

    model_params: QgParams
        An instance of the model parameters.
    delta_x: float, optional
        Spatial step in the zonal direction `x` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    delta_y: float, optional
        Spatial step in the meridional direction `y` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    dimensional: bool, optional
        Indicate if the output diagnostic must be dimensionalized or not.
        Default to `True`.

    Attributes
    ----------

    dimensional: bool
        Indicate if the output diagnostic must be dimensionalized or not.

    """

    def __init__(self, model_params, delta_x=None, delta_y=None, dimensional=True):

        AtmosphericStreamfunctionDiagnostic.__init__(self, model_params, delta_x, delta_y, dimensional)

        self._plot_title = r'Atmospheric $\psi_{\rm a}^3$ streamfunction'
        self._plot_units = r" (in " + self._model_params.get_variable_units(0) + r")"

    def _get_diagnostic(self, dimensional):

        if self._model_params.dynamic_T:
            offset = 1
        else:
            offset = 0

        vr = self._model_params.variables_range
        psi = np.swapaxes(self._data[:vr[0], ...].T @ np.swapaxes(self._grid_basis[offset:], 0, 1), 0, 1)
        theta = np.swapaxes(self._data[vr[0]+offset:vr[1], ...].T @ np.swapaxes(self._grid_basis[offset:], 0, 1), 0, 1)

        psi3 = psi - theta

        if dimensional:
            self._diagnostic_data = psi3 * self._model_params.streamfunction_scaling
            self._diagnostic_data_dimensional = True
        else:
            self._diagnostic_data = psi3
            self._diagnostic_data_dimensional = False
        return self._diagnostic_data


class UpperLayerAtmosphericStreamfunctionDiagnostic(AtmosphericStreamfunctionDiagnostic):
    """Diagnostic giving the upper layer atmospheric streamfunction fields :math:`\\psi^1_{\\rm a}`.
    Computed as :math:`\\psi^1_{\\rm a} = \\psi_{\\rm a} + \\theta_{\\rm a}` where :math:`\\psi_{\\rm a}` and :math:`\\theta_{\\rm a}` are
    respectively the barotropic and baroclinic streamfunctions.
    See also the :ref:`files/model/atmosphere:Atmospheric component` and :ref:`files/model/oro_model:Mid-layer equations and the thermal wind relation` sections.

    Parameters
    ----------

    model_params: QgParams
        An instance of the model parameters.
    delta_x: float, optional
        Spatial step in the zonal direction `x` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    delta_y: float, optional
        Spatial step in the meridional direction `y` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    dimensional: bool, optional
        Indicate if the output diagnostic must be dimensionalized or not.
        Default to `True`.

    Attributes
    ----------

    dimensional: bool
        Indicate if the output diagnostic must be dimensionalized or not.
    """

    def __init__(self, model_params, delta_x=None, delta_y=None, dimensional=True):

        AtmosphericStreamfunctionDiagnostic.__init__(self, model_params, delta_x, delta_y, dimensional)

        self._plot_title = r'Atmospheric $\psi_{\rm a}^1$ streamfunction'
        self._plot_units = r" (in " + self._model_params.get_variable_units(0) + r")"

    def _get_diagnostic(self, dimensional):

        if self._model_params.dynamic_T:
            offset = 1
        else:
            offset = 0

        vr = self._model_params.variables_range
        psi = np.swapaxes(self._data[:vr[0], ...].T @ np.swapaxes(self._grid_basis[offset:], 0, 1), 0, 1)
        theta = np.swapaxes(self._data[vr[0]+offset:vr[1], ...].T @ np.swapaxes(self._grid_basis[offset:], 0, 1), 0, 1)

        psi1 = psi + theta

        if dimensional:
            self._diagnostic_data = psi1 * self._model_params.streamfunction_scaling
            self._diagnostic_data_dimensional = True
        else:
            self._diagnostic_data = psi1
            self._diagnostic_data_dimensional = False
        return self._diagnostic_data


class MiddleAtmosphericStreamfunctionDiagnostic(AtmosphericStreamfunctionDiagnostic):
    """Diagnostic giving the middle atmospheric streamfunction fields :math:`\\psi_{\\rm a}` at 500hPa, i.e. the barotropic streamfunction of the system.
    See also :ref:`files/model/oro_model:Mid-layer equations and the thermal wind relation` sections.

    Parameters
    ----------

    model_params: QgParams
        An instance of the model parameters.
    delta_x: float, optional
        Spatial step in the zonal direction `x` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    delta_y: float, optional
        Spatial step in the meridional direction `y` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    dimensional: bool, optional
        Indicate if the output diagnostic must be dimensionalized or not.
        Default to `True`.
    geopotential: bool, optional
        Dimensionalize the field in geopotential height (in meter).
        Default to `False`.

    Attributes
    ----------

    dimensional: bool
        Indicate if the output diagnostic must be dimensionalized or not.
    """

    def __init__(self, model_params, delta_x=None, delta_y=None, dimensional=True, geopotential=False):

        AtmosphericStreamfunctionDiagnostic.__init__(self, model_params, delta_x, delta_y, dimensional)

        if geopotential:
            self._plot_title = r'Atmospheric 500hPa geopotential height'
            self._plot_units = r" (in m)"
        else:
            self._plot_title = r'Atmospheric 500hPa $\psi_{\rm a}$ streamfunction'
            self._plot_units = r" (in " + self._model_params.get_variable_units(0) + r")"
        self._default_plot_kwargs['cmap'] = plt.get_cmap('gist_rainbow_r')
        self.geopotential = geopotential
        if geopotential:
            self._color_bar_format = False

    def _get_diagnostic(self, dimensional):

        if self._model_params.dynamic_T:
            offset = 1
        else:
            offset = 0

        vr = self._model_params.variables_range
        psi = np.swapaxes(self._data[:vr[0], ...].T @ np.swapaxes(self._grid_basis[offset:], 0, 1), 0, 1)

        factor = 1.

        if dimensional:
            factor *= self._model_params.streamfunction_scaling
            self._diagnostic_data_dimensional = True
            if self.geopotential:
                factor *= self._model_params.geopotential_scaling
        else:
            self._diagnostic_data_dimensional = False
        self._diagnostic_data = psi * factor
        return self._diagnostic_data


class OceanicStreamfunctionDiagnostic(FieldDiagnostic):
    """General base class for oceanic streamfunction fields diagnostic.
    Provide a spatial gridded representation of the fields.

    Parameters
    ----------

    model_params: QgParams
        An instance of the model parameters.
    delta_x: float, optional
        Spatial step in the zonal direction `x` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    delta_y: float, optional
        Spatial step in the meridional direction `y` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    dimensional: bool
        Indicate if the output diagnostic must be dimensionalized or not.

    Attributes
    ----------

    dimensional: bool
        Indicate if the output diagnostic must be dimensionalized or not.
    """

    def __init__(self, model_params, delta_x=None, delta_y=None, dimensional=True):

        FieldDiagnostic.__init__(self, model_params, dimensional)
        self._configure(delta_x=delta_x, delta_y=delta_y)

        self._default_plot_kwargs['cmap'] = plt.get_cmap('jet')

    def _compute_grid(self, delta_x=None, delta_y=None):

        if delta_x is None:
            oms = self._model_params.oblocks
            if oms is None:
                warnings.warn("OceanicStreamfunctionDiagnostic: Unable to configure the grid automatically. Oceanic wavenumbers information not " +
                              "present in the model's parameters ! Please call the compute_grid method with the delta_x and delta_y parameters.")
                return 1
            xwn = [oms[i][0] for i in range(len(oms))]
            mxwn = max(xwn)
            n_point_x = 4 * mxwn + 2
        else:
            n_point_x = int(np.ceil((2 * np.pi / self._model_params.scale_params.n) / delta_x) + 1)

        if delta_y is None:
            oms = self._model_params.oblocks
            if oms is None:
                warnings.warn("OceanicStreamfunctionDiagnostic: Unable to configure the grid automatically. Oceanic wavenumbers information not " +
                              "present in the model's parameters ! Please call the compute_grid method with the delta_x and delta_y parameters.")
                return 1
            ywn = [oms[i][1] for i in range(len(oms))]
            mywn = max(ywn)
            n_point_y = 4 * mywn + 2
        else:
            n_point_y = int(np.ceil(np.pi / delta_y) + 1)

        x = np.linspace(0., 2 * np.pi / self._model_params.scale_params.n, n_point_x)
        y = np.linspace(0., np.pi, n_point_y)
        self._X, self._Y = np.meshgrid(x, y)

    def _configure(self, model_params=None, delta_x=None, delta_y=None):

        if not self._ocean:
            warnings.warn("OceanicStreamfunctionDiagnostic: No ocean configuration found in the provided parameters. This model version does not have an ocean. " +
                          "Please check your configuration.")
            return 1

        self._compute_grid(delta_x, delta_y)
        basis = self._model_params.oceanic_basis

        self._grid_basis = create_grid_basis(basis, self._X, self._Y, self._subs)


class OceanicLayerStreamfunctionDiagnostic(OceanicStreamfunctionDiagnostic):
    """Diagnostic giving the oceanic layer streamfunction fields :math:`\\psi_{\\rm o}`.

    Parameters
    ----------

    model_params: QgParams
        An instance of the model parameters.
    delta_x: float, optional
        Spatial step in the zonal direction `x` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    delta_y: float, optional
        Spatial step in the meridional direction `y` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    dimensional: bool, optional
        Indicate if the output diagnostic must be dimensionalized or not.
        Default to `True`.
    conserved: bool, optional
        Whether to plot the conserved oceanic fields or not. Default to `True`.

    Attributes
    ----------

    dimensional: bool
        Indicate if the output diagnostic must be dimensionalized or not.
    """

    def __init__(self, model_params, delta_x=None, delta_y=None, dimensional=True, conserved=True):

        OceanicStreamfunctionDiagnostic.__init__(self, model_params, delta_x, delta_y, dimensional)

        vr = self._model_params.variables_range
        self._plot_title = r'Oceanic streamfunction'
        self._plot_units = r" (in " + self._model_params.get_variable_units(vr[1]) + r")"

        self._conserved = conserved

        self._fields_average = list()
        basis = self._model_params.oceanic_basis

        for func in basis.num_functions(self._subs):
            average = dblquad(func, 0, np.pi, 0, 2*np.pi/model_params.scale_params.n)[0] / (np.pi * 2*np.pi/model_params.scale_params.n)
            self._fields_average.append(average)

        self._fields_average = np.array(self._fields_average)

    def _get_diagnostic(self, dimensional):

        if self._model_params.dynamic_T:
            offset = 1
        else:
            offset = 0

        vr = self._model_params.variables_range
        if self._conserved:
            cgrid = np.swapaxes(np.swapaxes(self._grid_basis, 0, -1) - self._fields_average, 0, -1)
        else:
            cgrid = self._grid_basis
        psi = np.swapaxes(self._data[vr[1]:vr[2], ...].T @ np.swapaxes(cgrid[offset:], 0, 1), 0, 1)

        if dimensional:
            self._diagnostic_data = psi * self._model_params.streamfunction_scaling
            self._diagnostic_data_dimensional = True
        else:
            self._diagnostic_data = psi
            self._diagnostic_data_dimensional = False
        return self._diagnostic_data


if __name__ == '__main__':
    from qgs.params.params import QgParams
    from qgs.params.params import QgParams
    from qgs.integrators.integrator import RungeKuttaIntegrator
    from qgs.functions.tendencies import create_tendencies

    pars = QgParams()
    pars.set_atmospheric_channel_fourier_modes(2, 2)
    f, Df = create_tendencies(pars)
    integrator = RungeKuttaIntegrator()
    integrator.set_func(f)
    ic = np.random.rand(pars.ndim) * 0.1
    integrator.integrate(0., 200000., 0.1, ic=ic, write_steps=5)
    time, traj = integrator.get_trajectories()
    integrator.terminate()

    psi3 = LowerLayerAtmosphericStreamfunctionDiagnostic(pars)
    psi3(time, traj)

    psi1 = UpperLayerAtmosphericStreamfunctionDiagnostic(pars)
    psi1(time, traj)

    psi = MiddleAtmosphericStreamfunctionDiagnostic(pars)
    psi(time, traj)



"""
    Diagnostic eddy classes
    =======================

    Classes defining eddy related fields diagnostics.

    Description of the classes
    --------------------------

    * :class:`MiddleAtmosphericEddyHeatFluxDiagnostic`: Diagnostic giving the middle atmospheric eddy heat flux field.
    * :class:`MiddleAtmosphericEddyHeatFluxProfileDiagnostic`: Diagnostic giving the middle atmospheric eddy heat flux zonally averaged profile.

"""

import warnings

import numpy as np
from scipy.integrate import simpson
import matplotlib.pyplot as plt

from qgs.diagnostics.base import FieldDiagnostic, ProfileDiagnostic
from qgs.diagnostics.temperatures import MiddleAtmosphericTemperatureAnomalyDiagnostic
from qgs.diagnostics.wind import MiddleAtmosphericVWindDiagnostic


class MiddleAtmosphericEddyHeatFluxDiagnostic(FieldDiagnostic):
    """Diagnostic giving the middle atmospheric eddy heat flux field.
    Computed as :math:`v'_{\\rm a} \\, T'_{\\rm a}` and scaled with the
    atmospheric specific heat capicity if available (through the `heat_capacity` argument or the
    :attr:`~.AtmosphericTemperatureParams.gamma` parameter).

    Parameters
    ----------

    model_params: QgParams
        An instance of the model parameters.
    delta_x: float, optional
        Spatial step in the zonal direction `x` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    delta_y: float, optional
        Spatial step in the meridional direction `y` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    dimensional: bool, optional
        Indicate if the output diagnostic must be dimensionalized or not.
        Default to `True`.
    temp_mean_state: MiddleAtmosphericTemperatureDiagnostic, optional
        A temperature diagnostic with a long trajectory as data to compute the mean temperature field.
        If not provided, compute the mean with the data stored in the object.
    vwind_mean_state: MiddleAtmosphericVWindDiagnostic, optional
        A :math:`v` wind diagnostic with a long trajectory as data to compute the mean wind field.
        If not provided, compute the mean with the data stored in the object.
    heat_capacity: float, optional
        The air specific heat capacity. If not provided, uses the one of :attr:`~.AtmosphericTemperatureParams.gamma` if
        available or or let the heat flux in K m s^{-1}.

    Attributes
    ----------

    dimensional: bool
        Indicate if the output diagnostic must be dimensionalized or not.

    """

    def __init__(self, model_params, delta_x=None, delta_y=None, dimensional=True, temp_mean_state=None, vwind_mean_state=None, heat_capacity=None):

        FieldDiagnostic.__init__(self, model_params, dimensional)

        self._plot_title = r'Eddy heat flux in the middle of the atmosphere'
        if heat_capacity is not None or model_params.atemperature_params.gamma is not None:
            self._plot_title += r" $\gamma_{\rm a} v'_{\rm a} \, T'_{\rm a}$"
            self._plot_units = r" (in " + r'W m$^{-1}$' + r")"
        else:
            self._plot_title += r" $v'_{\rm a} \, T'_{\rm a}$"
            self._plot_units = r" (in " + r'K m s$^{-1}$' + r")"
        self._default_plot_kwargs['cmap'] = plt.get_cmap('magma')
        self._color_bar_format = False

        self._tdiag = MiddleAtmosphericTemperatureAnomalyDiagnostic(model_params, delta_x, delta_y, dimensional)
        self._vdiag = MiddleAtmosphericVWindDiagnostic(model_params, delta_x, delta_y, dimensional)

        self._X = self._tdiag._X
        self._Y = self._tdiag._Y

        self._temp_mean_state = temp_mean_state
        self._vwind_mean_state = vwind_mean_state

        self._heat_capacity = heat_capacity

    def _compute_grid(self, delta_x=None, delta_y=None):
        pass

    def _configure(self, delta_x=None, delta_y=None):
        pass

    def _get_diagnostic(self, dimensional):

        self._tdiag.set_data(self._time, self._data)
        self._vdiag.set_data(self._time, self._data)

        T = self._tdiag._get_diagnostic(dimensional)
        V = self._vdiag._get_diagnostic(dimensional)

        if self._temp_mean_state is not None:
            Tmean = self._temp_mean_state._get_diagnostic(dimensional).mean(axis=0)
        else:
            Tmean = np.mean(T, axis=0)
        if self._vwind_mean_state is not None:
            Vmean = self._vwind_mean_state._get_diagnostic(dimensional).mean(axis=0)
        else:
            Vmean = np.mean(V, axis=0)

        self._diagnostic_data = (T - Tmean) * (V - Vmean)
        if dimensional:
            if self._model_params.atemperature_params.gamma is not None:
                self._diagnostic_data = self._diagnostic_data * self._model_params.atemperature_params.gamma
            elif self._heat_capacity is not None:
                self._diagnostic_data = self._diagnostic_data * self._heat_capacity
            self._diagnostic_data_dimensional = True
        else:
            self._diagnostic_data_dimensional = False
        return self._diagnostic_data


class MiddleAtmosphericEddyHeatFluxProfileDiagnostic(ProfileDiagnostic):
    """Diagnostic giving the middle atmospheric eddy heat flux zonally averaged profile.
    Computed as :math:`\\Phi_{\\rm e} = \\overline{v'_{\\rm a} \\, T'_{\\rm a}} = \\frac{n}{2\\pi} \\, \\int_0^{2\\pi/n} \\Phi_{\\rm e} \\, \\mathrm{d} x` where
    :math:`v'_{\\rm a} \\, T'_{\\rm a}` is the eddy heat flux scaled with the
    atmospheric specific heat capicity if available (through the `heat_capacity` argument or the
    :attr:`~.AtmosphericTemperatureParams.gamma` parameter).

    Parameters
    ----------

    model_params: QgParams
        An instance of the model parameters.
    delta_x: float, optional
        Spatial step in the zonal direction `x` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    delta_y: float, optional
        Spatial step in the meridional direction `y` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    dimensional: bool, optional
        Indicate if the output diagnostic must be dimensionalized or not.
        Default to `True`.
    temp_mean_state: MiddleAtmosphericTemperatureDiagnostic, optional
        A temperature diagnostic with a long trajectory as data to compute the mean temperature field.
        If not provided, compute the mean with the data stored in the object.
    vwind_mean_state: MiddleAtmosphericVWindDiagnostic, optional
        A :math:`v` wind diagnostic with a long trajectory as data to compute the mean wind field.
        If not provided, compute the mean with the data stored in the object.
    heat_capacity: float, optional
        The air specific heat capacity. If not provided, uses the one of :attr:`~.AtmosphericTemperatureParams.gamma` if
        available or or let the heat flux in K m s^{-1}.

    Attributes
    ----------

    dimensional: bool
        Indicate if the output diagnostic must be dimensionalized or not.

    """

    def __init__(self, model_params, delta_x=None, delta_y=None, dimensional=True, temp_mean_state=None, vwind_mean_state=None, heat_capacity=None):

        ProfileDiagnostic.__init__(self, model_params, dimensional)

        self._flux = MiddleAtmosphericEddyHeatFluxDiagnostic(model_params, delta_x, delta_y, dimensional, temp_mean_state, vwind_mean_state, heat_capacity)
        self._plot_title = r'Zonally averaged profile'
        self._plot_label = r'Middle atmospheric eddy heat flux'
        if heat_capacity is not None or model_params.atemperature_params.gamma is not None:
            self._plot_label += r" $\gamma_{\rm a} \overline{v'_{\rm a} \, T'_{\rm a}}$"
            self._plot_units = r'W m$^{-1}$'
        else:
            self._plot_label += r" $\overline{v'_{\rm a} \, T'_{\rm a}}$"
            self._plot_units = r'K m s$^{-1}$'
        self._axis_label = r'$y$'
        self._configure()

    def _configure(self):
        self._points_coordinates = self._flux._Y[:, 0]

    def _get_diagnostic(self, dimensional):

        self._flux.set_data(self._time, self._data)

        flux = self._flux._get_diagnostic(dimensional)
        dX = self._flux._X[0, 1] - self._flux._X[0, 0]

        iflux = simpson(flux, dx=dX, axis=2) / (2*np.pi / self._model_params.scale_params.n)

        self._diagnostic_data = iflux
        if dimensional:
            self._diagnostic_data_dimensional = True
        else:
            self._diagnostic_data_dimensional = False
        return self._diagnostic_data


if __name__ == '__main__':
    from qgs.params.params import QgParams
    from qgs.integrators.integrator import RungeKuttaIntegrator
    from qgs.functions.tendencies import create_tendencies

    pars = QgParams()
    pars.set_atmospheric_channel_fourier_modes(2, 2)
    f, Df = create_tendencies(pars)
    integrator = RungeKuttaIntegrator()
    integrator.set_func(f)
    ic = np.random.rand(pars.ndim) * 0.1
    integrator.integrate(0., 200000., 0.1, ic=ic, write_steps=5)
    time, traj = integrator.get_trajectories()
    integrator.terminate()

    flux = MiddleAtmosphericEddyHeatFluxDiagnostic(pars)
    flux.set_data(time, traj)

    iflux = MiddleAtmosphericEddyHeatFluxProfileDiagnostic(pars, delta_x=0.25, delta_y=0.15)
    iflux.set_data(time, traj)



"""
    Symbolic matrix operation module
    ================================

    This module supports operations and functions on the symbolic sparse tensors defined in
    the :class:`~.tensors.symbolic_qgtensor.SymbolicQgsTensor` objects.

"""

from sympy import tensorproduct, tensorcontraction
from qgs.functions.util import add_to_dict


def symbolic_tensordot(a, b, axes=2):
    """Compute tensor dot product along specified axes of two sympy symbolic arrays

    This is based on `Numpy`_ :meth:`~numpy.tensordot` .

    .. _Numpy: https://numpy.org/

    Parameters
    ----------
    a, b: ~sympy.tensor.array.DenseNDimArray or ~sympy.tensor.array.SparseNDimArray
        Arrays to take the dot product of.

    axes: int
        Sum over the last `axes` axes of `a` and the first `axes` axes
        of `b` in order. The sizes of the corresponding axes must match.

    Returns
    -------
    output: Sympy tensor
        The tensor dot product of the input.

    """
    as_ = a.shape
    nda = len(as_)
    
    a_com = [nda+i for i in range(-axes, 0)]
    b_com = [nda+i for i in range(axes)]
    sum_cols = tuple(a_com + b_com)
    
    prod = tensorproduct(a, b)
    
    return tensorcontraction(prod, sum_cols)


def symbolic_sparse_mult2(dic, vec_a):
    """Symbolic multiplication of a tensor with one vector:
    :math:`A_{i,j} = {\\displaystyle \\sum_{k=0}^{\\mathrm{ndim}}} \\, \\mathcal{T}_{i,j,k} \\, a_k`

    Parameters
    ----------
    dic: dict(~sympy.core.symbol.Symbol)
        A dictionary whose keys are the coordinates of the tensor, and the dictionary values are the values of the
        tensor.
    vec_a: list(~sympy.core.symbol.Symbol)
        The list :math:`a_k` to contract the tensor with entries of Sympy Symbols.
        Must be of shape (:attr:`~.params.QgParams.ndim` + 1,).

    Returns
    -------
    res: dict(~sympy.core.symbol.Symbol)
        The matrix :math:`A_{i,j}`, of shape (:attr:`~.params.QgParams.ndim` + 1, :attr:`~.params.QgParams.ndim` + 1),
        contained in a dictionary, where the keys are the tensor coordinates, and the values are the tensor values.
    """
    res = dict()

    for key in dic.keys():
        coo1, coo2, coo3 = key
        if coo1 > 0 and coo2 > 0:
            val = vec_a[coo3] * dic[key]
            res = add_to_dict(res, (coo1, coo2), val)

    return res


def symbolic_sparse_mult3(dic, vec_a, vec_b):
    """Symbolic multiplication of a tensor with two vectors:
    :math:`v_i = {\\displaystyle \\sum_{j,k=0}^{\\mathrm{ndim}}} \\, \\mathcal{T}_{i,j,k} \\, a_j \\, b_k`

    Parameters
    ----------
    dic: dict(~sympy.core.symbol.Symbol)
        A dictionary whose keys are the coordinates of the tensor, and the dictionary values are the values of the
        tensor.
    vec_a: list(~sympy.core.symbol.Symbol)
        The list :math:`a_j` to contract the tensor with entries of Sympy Symbols.
        Must be of shape (:attr:`~.params.QgParams.ndim` + 1,).
    vec_b: list(~sympy.core.symbol.Symbol)
        The list :math:`b_k` to contract the tensor with entries of Sympy Symbols.
        Must be of shape (:attr:`~.params.QgParams.ndim` + 1,).


    Returns
    -------
    res: dict(~sympy.core.symbol.Symbol)
        The vector :math:`v_i`, of shape (:attr:`~.params.QgParams.ndim` + 1,), contained in a dictionary, where
        the keys are the tensor coordinates, and the values are the tensor values.

    """
    res = dict()

    for key in dic.keys():
        coo1, coo2, coo3 = key
        if coo1 > 0:
            val = vec_a[coo2] * vec_b[coo3] * dic[key]
            res = add_to_dict(res, coo1, val)

    return res


def symbolic_sparse_mult4(dic, vec_a, vec_b, vec_c):
    """Symbolic multiplication of a rank-5 tensor with three vectors:
    :math:`A_{i, j} = {\\displaystyle \\sum_{k,l,m=0}^{\\mathrm{ndim}}} \\, \\mathcal{T}_{i,j,k,l, m} \\, a_k \\, b_l \\, c_m`


    Parameters
    ----------
    dic: dict(~sympy.core.symbol.Symbol)
        A dictionary where they keys are a tuple of 5 elements which are the coordinates of the tensor values,
        which are contained in the dictionary values.
    vec_a: list(~sympy.core.symbol.Symbol)
        The list :math:`a_j` to contract the tensor with entries of Sympy Symbols.
        Must be of shape (:attr:`~.params.QgParams.ndim` + 1,).
    vec_b: list(~sympy.core.symbol.Symbol)
        The list :math:`b_k` to contract the tensor with entries of Sympy Symbols.
        Must be of shape (:attr:`~.params.QgParams.ndim` + 1,).
    vec_c: list(~sympy.core.symbol.Symbol)
        The list :math:`c_l` to contract the tensor with entries of Sympy Symbols.
        Must be of shape (:attr:`~.params.QgParams.ndim` + 1,).

    Returns
    -------
    res: dict(~sympy.core.symbol.Symbol)
        The matrix :math:`A_{i, j}`, of shape (:attr:`~.params.QgParams.ndim` + 1, :attr:`~.params.QgParams.ndim` + 1),
        contained in a dictionary, where the keys are the tensor coordinates, and the values are the tensor values.
    """
    res = dict()

    for key in dic.keys():
        coo1, coo2, coo3, coo4, coo5 = key
        if coo1 > 0 and coo2 > 0:
            val = vec_a[coo3] * vec_b[coo4] * vec_c[coo5] * dic[key]
            res = add_to_dict(res, (coo1, coo2), val)

    return res


def symbolic_sparse_mult5(dic, vec_a, vec_b, vec_c, vec_d):
    """Symbolic multiplication of a rank-5 tensor with four vectors:
    :math:`v_i = {\\displaystyle \\sum_{j,k,l,m=0}^{\\mathrm{ndim}}} \\, \\mathcal{T}_{i,j,k,l,m} \\, a_j \\, b_k \\, c_l \\, d_m`

    Parameters
    ----------
    dic: dict(~sympy.core.symbol.Symbol)
        A dictionary whose keys are the coordinates of the tensor, and the dictionary values are the values of the
        tensor.
    vec_a: list(~sympy.core.symbol.Symbol)
        The list :math:`a_j` to contract the tensor with entries of Sympy Symbols.
        Must be of shape (:attr:`~.params.QgParams.ndim` + 1,).
    vec_b: list(~sympy.core.symbol.Symbol)
        The list :math:`b_k` to contract the tensor with entries of Sympy Symbols.
        Must be of shape (:attr:`~.params.QgParams.ndim` + 1,).
    vec_c: list(~sympy.core.symbol.Symbol)
        The list :math:`c_l` to contract the tensor with entries of Sympy Symbols.
        Must be of shape (:attr:`~.params.QgParams.ndim` + 1,).
    vec_d: list(~sympy.core.symbol.Symbol)
        The list :math:`d_m` to contract the tensor with entries of Sympy Symbols.
        Must be of shape (:attr:`~.params.QgParams.ndim` + 1,).

    Returns
    -------
    res: dict(~sympy.core.symbol.Symbol)
        The vector :math:`v_i`, of shape (:attr:`~.params.QgParams.ndim` + 1,), contained in a dictionary,
        where the keys are the tensor coordinates, and the values are the tensor values.
    """
    res = dict()

    for key in dic.keys():
        coo1, coo2, coo3, coo4, coo5 = key
        if coo1 > 0:
            val = vec_a[coo2] * vec_b[coo3] * vec_c[coo4] * vec_d[coo5] * dic[key]
            res = add_to_dict(res, coo1, val)

    return res




"""
    Tendencies definition module
    ============================

    This module provides functions to create the tendencies functions of the model, based on
    its parameters.

"""
import numpy as np
from numba import njit

from qgs.inner_products.analytic import AtmosphericAnalyticInnerProducts, OceanicAnalyticInnerProducts, GroundAnalyticInnerProducts
from qgs.inner_products.symbolic import AtmosphericSymbolicInnerProducts, OceanicSymbolicInnerProducts, GroundSymbolicInnerProducts
from qgs.tensors.qgtensor import QgsTensor, QgsTensorDynamicT, QgsTensorT4
from qgs.tensors.atmo_thermo_tensor import AtmoThermoTensor, AtmoThermoTensorDynamicT, AtmoThermoTensorT4
from qgs.functions.sparse_mul import sparse_mul5, sparse_mul4, sparse_mul3, sparse_mul2


def create_tendencies(params, return_inner_products=False, return_qgtensor=False):
    """Function to handle the inner products and tendencies tensors construction.
    Returns the tendencies function :math:`\\boldsymbol{f}` determining the model's ordinary differential
    equations:

    .. math:: \dot{\\boldsymbol{x}} = \\boldsymbol{f}(\\boldsymbol{x})

    which is for the model's integration.

    It returns also the linearized tendencies
    :math:`\\boldsymbol{\mathrm{J}} \equiv \\boldsymbol{\mathrm{D}f} = \\frac{\partial \\boldsymbol{f}}{\partial \\boldsymbol{x}}`
    (Jacobian matrix) which are used by the tangent linear model:

    .. math :: \dot{\\boldsymbol{\delta x}} = \\boldsymbol{\mathrm{J}}(\\boldsymbol{x}) \cdot \\boldsymbol{\delta x}

    Parameters
    ----------
    params: QgParams
        The parameters fully specifying the model configuration.
    return_inner_products: bool
        If True, return the inner products of the model. Default to False.
    return_qgtensor: bool
        If True, return the tendencies tensor of the model. Default to False.


    Returns
    -------
    f: callable
        The numba-jitted tendencies function.
    Df: callable
        The numba-jitted linearized tendencies function.
    inner_products: (AtmosphericInnerProducts, OceanicInnerProducts)
        If `return_inner_products` is True, the inner products of the system.
    qgtensor: QgsTensor
        If `return_qgtensor` is True, the tendencies tensor of the system.
    """

    if params.ablocks is not None:
        aip = AtmosphericAnalyticInnerProducts(params)
    elif params.atmospheric_basis is not None:
        aip = AtmosphericSymbolicInnerProducts(params)
    else:
        aip = None

    if params.oblocks is not None:
        oip = OceanicAnalyticInnerProducts(params)
    elif params.oceanic_basis is not None:
        oip = OceanicSymbolicInnerProducts(params)
    else:
        oip = None

    if params.gblocks is not None:
        gip = GroundAnalyticInnerProducts(params)
    elif params.ground_basis is not None:
        gip = GroundSymbolicInnerProducts(params)
    else:
        gip = None

    if aip is not None and oip is not None:
        if not aip.connected_to_ocean:
            aip.connect_to_ocean(oip)
    elif aip is not None and gip is not None:
        if not aip.connected_to_ground:
            aip.connect_to_ground(gip)

    if params.T4:
        agotensor = QgsTensorT4(params, aip, oip, gip)
    elif params.dynamic_T:
        agotensor = QgsTensorDynamicT(params, aip, oip, gip)
    else:
        agotensor = QgsTensor(params, aip, oip, gip)

    coo = agotensor.tensor.coords.T
    val = agotensor.tensor.data

    jcoo = agotensor.jacobian_tensor.coords.T
    jval = agotensor.jacobian_tensor.data

    if params.T4 or params.dynamic_T:
        @njit
        def f(t, x):
            xx = np.concatenate((np.full((1,), 1.), x))
            xr = sparse_mul5(coo, val, xx, xx, xx, xx)
            return xr[1:]

        @njit
        def Df(t, x):
            xx = np.concatenate((np.full((1,), 1.), x))
            mul_jac = sparse_mul4(jcoo, jval, xx, xx, xx)
            return mul_jac[1:, 1:]
    else:
        @njit
        def f(t, x):
            xx = np.concatenate((np.full((1,), 1.), x))
            xr = sparse_mul3(coo, val, xx, xx)
            return xr[1:]

        @njit
        def Df(t, x):
            xx = np.concatenate((np.full((1,), 1.), x))
            mul_jac = sparse_mul2(jcoo, jval, xx)
            return mul_jac[1:, 1:]

    ret = list()
    ret.append(f)
    ret.append(Df)
    if return_inner_products:
        ret.append((aip, oip, gip))
    if return_qgtensor:
        ret.append(agotensor)
    return ret


def create_atmo_thermo_tendencies(params, return_atmo_thermo_tensor=False):
    """Returns a function which return a part of the atmospheric thermodynamic tendencies useful for computing the vertical wind.

    Parameters
    ----------
    params: QgParams
        The parameters fully specifying the model configuration.
    return_atmo_thermo_tensor: bool
        If True, return the tendencies tensor of these tencencies. Default to False.


    Returns
    -------
    f: callable
        The numba-jitted tendencies function.
    atmo_thermo_tensor: AtmoThermoTensor
        If `return_atmo_thermo_tensor` is True, the tendencies tensor of the system.
    """

    if params.ablocks is not None:
        aip = AtmosphericAnalyticInnerProducts(params)
    elif params.atmospheric_basis is not None:
        aip = AtmosphericSymbolicInnerProducts(params)
    else:
        aip = None

    if params.oblocks is not None:
        oip = OceanicAnalyticInnerProducts(params)
    elif params.oceanic_basis is not None:
        oip = OceanicSymbolicInnerProducts(params)
    else:
        oip = None

    if params.gblocks is not None:
        gip = GroundAnalyticInnerProducts(params)
    elif params.ground_basis is not None:
        gip = GroundSymbolicInnerProducts(params)
    else:
        gip = None

    if aip is not None and oip is not None:
        if not aip.connected_to_ocean:
            aip.connect_to_ocean(oip)
    elif aip is not None and gip is not None:
        if not aip.connected_to_ground:
            aip.connect_to_ground(gip)

    if params.T4:
        agotensor = AtmoThermoTensorT4(params, aip, oip, gip)
    elif params.dynamic_T:
        agotensor = AtmoThermoTensorDynamicT(params, aip, oip, gip)
    else:
        agotensor = AtmoThermoTensor(params, aip, oip, gip)

    coo = agotensor.tensor.coords.T
    val = agotensor.tensor.data

    if params.T4 or params.dynamic_T:
        @njit
        def f(t, x):
            xx = np.concatenate((np.full((1,), 1.), x))
            xr = sparse_mul5(coo, val, xx, xx, xx, xx)
            return xr[1:]

    else:
        @njit
        def f(t, x):
            xx = np.concatenate((np.full((1,), 1.), x))
            xr = sparse_mul3(coo, val, xx, xx)
            return xr[1:]

    if return_atmo_thermo_tensor:
        ret = list()
        ret.append(f)
        ret.append(agotensor)
    else:
        ret = f

    return ret


if __name__ == '__main__':
    from qgs.params.params import QgParams

    params = QgParams()
    params.set_atmospheric_channel_fourier_modes(2, 2, mode='symbolic')
    params.set_oceanic_basin_fourier_modes(2, 4, mode='symbolic')
    f, Df = create_tendencies(params)



"""
    Sparse matrix operation module
    ==============================

    This module supports operations and functions on the sparse tensors defined in
    the :class:`~.tensors.qgtensor.QgsTensor` objects.

"""
import numpy as np
from numba import njit


@njit
def sparse_mul2(coo, value, vec):
    """Sparse multiplication of a tensor with one vector:
    :math:`A_{i,j} = {\\displaystyle \\sum_{k=0}^{\\mathrm{ndim}}} \\, \\mathcal{T}_{i,j,k} \\, a_k`

    Warnings
    --------
    It is a Numba-jitted function, so it cannot take a :class:`sparse.COO` sparse tensor directly.
    The tensor coordinates list and values must be provided separately by the user.

    In principle, this will be solved later in `sparse`, see https://github.com/pydata/sparse/issues/378.

    Parameters
    ----------
    coo: ~numpy.ndarray(int)
        A 2D array of shape (n_elems, 3), a list of n_elems tensor coordinates corresponding to each value provided.
    value: ~numpy.ndarray(float)
        A 1D array of shape (n_elems,), a list of value in the tensor
    vec: ~numpy.ndarray(float)
        The vector :math:`a_k` to contract the tensor with. Must be of shape (:attr:`~.params.QgParams.ndim` + 1,).

    Returns
    -------
    ~numpy.ndarray(float)
        The matrix :math:`A_{i,j}`, of shape (:attr:`~.params.QgParams.ndim` + 1, :attr:`~.params.QgParams.ndim` + 1).
    """

    res = np.zeros((len(vec), len(vec)))

    for n in range(coo.shape[0]):
        res[coo[n, 0], coo[n, 1]] += vec[coo[n, 2]] * value[n]

    return res


@njit
def sparse_mul3(coo, value, vec_a, vec_b):
    """Sparse multiplication of a tensor with two vectors:
    :math:`v_i = {\\displaystyle \\sum_{j,k=0}^{\\mathrm{ndim}}} \\, \\mathcal{T}_{i,j,k} \\, a_j \\, b_k`

    Warnings
    --------
    It is a Numba-jitted function, so it cannot take a :class:`sparse.COO` sparse tensor directly.
    The tensor coordinates list and values must be provided separately by the user.

    In principle, this will be solved later in `sparse`, see https://github.com/pydata/sparse/issues/378.

    Parameters
    ----------
    coo: ~numpy.ndarray(int)
        A 2D array of shape (n_elems, 3), a list of n_elems tensor coordinates corresponding to each value provided.
    value: ~numpy.ndarray(float)
        A 1D array of shape (n_elems,), a list of value in the tensor
    vec_a: ~numpy.ndarray(float)
        The vector :math:`a_j` to contract the tensor with. Must be of shape (:attr:`~.params.QgParams.ndim` + 1,).
    vec_b: ~numpy.ndarray(float)
        The vector :math:`b_k` to contract the tensor with. Must be of shape (:attr:`~.params.QgParams.ndim` + 1,).

    Returns
    -------
    ~numpy.ndarray(float)
        The vector :math:`v_i`, of shape (:attr:`~.params.QgParams.ndim` + 1,).
    """
    res = np.zeros_like(vec_a)
    n_elems = coo.shape[0]
    for n in range(n_elems):
        res[coo[n, 0]] += vec_a[coo[n, 1]] * vec_b[coo[n, 2]] * value[n]
    res[0] = 1.
    return res


@njit
def sparse_mul4(coo, value, vec_a, vec_b, vec_c):
    """Sparse multiplication of a rank-5 tensor with three vectors:
    :math:`A_{i, j} = {\\displaystyle \\sum_{k,l,m=0}^{\\mathrm{ndim}}} \\, \\mathcal{T}_{i,j,k,l, m} \\, a_k \\, b_l \\, c_m`

    Warnings
    --------
    It is a Numba-jitted function, so it cannot take a :class:`sparse.COO` sparse tensor directly.
    The tensor coordinates list and values must be provided separately by the user.

    In principle, this will be solved later in `sparse`, see https://github.com/pydata/sparse/issues/378.

    Parameters
    ----------
    coo: ~numpy.ndarray(int)
        A 2D array of shape (n_elems, 5), a list of n_elems tensor coordinates corresponding to each value provided.
    value: ~numpy.ndarray(float)
        A 1D array of shape (n_elems,), a list of value in the tensor
    vec_a: ~numpy.ndarray(float)
        The vector :math:`a_j` to contract the tensor with. Must be of shape (:attr:`~.params.QgParams.ndim` + 1,).
    vec_b: ~numpy.ndarray(float)
        The vector :math:`b_k` to contract the tensor with. Must be of shape (:attr:`~.params.QgParams.ndim` + 1,).
    vec_c: ~numpy.ndarray(float)
        The vector :math:`c_l` to contract the tensor with. Must be of shape (:attr:`~.params.QgParams.ndim` + 1,).

    Returns
    -------
    ~numpy.ndarray(float)
        The matrix :math:`A_{i, j}`, of shape (:attr:`~.params.QgParams.ndim` + 1, :attr:`~.params.QgParams.ndim` + 1).
    """
    res = np.zeros((len(vec_a), len(vec_a)))
    n_elems = coo.shape[0]
    for n in range(n_elems):
        res[coo[n, 0], coo[n, 1]] += vec_a[coo[n, 2]] * vec_b[coo[n, 3]] * vec_c[coo[n, 4]] * value[n]
    return res


@njit
def sparse_mul5(coo, value, vec_a, vec_b, vec_c, vec_d):
    """Sparse multiplication of a rank-5 tensor with four vectors:
    :math:`v_i = {\\displaystyle \\sum_{j,k,l,m=0}^{\\mathrm{ndim}}} \\, \\mathcal{T}_{i,j,k,l,m} \\, a_j \\, b_k \\, c_l \\, d_m`

    Warnings
    --------
    It is a Numba-jitted function, so it cannot take a :class:`sparse.COO` sparse tensor directly.
    The tensor coordinates list and values must be provided separately by the user.

    In principle, this will be solved later in `sparse`, see https://github.com/pydata/sparse/issues/378.

    Parameters
    ----------
    coo: ~numpy.ndarray(int)
        A 2D array of shape (n_elems, 5), a list of n_elems tensor coordinates corresponding to each value provided.
    value: ~numpy.ndarray(float)
        A 1D array of shape (n_elems,), a list of value in the tensor
    vec_a: ~numpy.ndarray(float)
        The vector :math:`a_j` to contract the tensor with. Must be of shape (:attr:`~.params.QgParams.ndim` + 1,).
    vec_b: ~numpy.ndarray(float)
        The vector :math:`b_k` to contract the tensor with. Must be of shape (:attr:`~.params.QgParams.ndim` + 1,).
    vec_c: ~numpy.ndarray(float)
        The vector :math:`c_l` to contract the tensor with. Must be of shape (:attr:`~.params.QgParams.ndim` + 1,).
    vec_d: ~numpy.ndarray(float)
        The vector :math:`d_m` to contract the tensor with. Must be of shape (:attr:`~.params.QgParams.ndim` + 1,).

    Returns
    -------
    ~numpy.ndarray(float)
        The vector :math:`v_i`, of shape (:attr:`~.params.QgParams.ndim` + 1,).
    """
    res = np.zeros_like(vec_a)
    n_elems = coo.shape[0]
    for n in range(n_elems):
        res[coo[n, 0]] += vec_a[coo[n, 1]] * vec_b[coo[n, 2]] * vec_c[coo[n, 3]] * vec_d[coo[n, 4]] * value[n]
    res[0] = 1.
    return res



"""
    Utility functions module
    ========================

    This module has some useful functions for the model.

"""


import numpy as np
from numba import njit


def add_to_dict(dic, loc, value):
    """Adds `value` to dictionary `dic`, with the dictionary key of `loc`.
    If the dictionary did not have a key of `loc` before, a new key is made.

    Parameters
    ----------
    dic: dict
        Dictionary to add the value to.
    loc:
        Item of the dictionary to add the value to.
    value:
        Value to add.
    """
    try:
        dic[loc] += value
    except:
        dic[loc] = value
    return dic


@njit
def reverse(a):
    """Numba-jitted function to reverse a 1D array.

    Parameters
    ----------
    a: ~numpy.ndarray
        The 1D array to reverse.

    Returns
    -------
    ~numpy.ndarray
        The reversed array.
    """
    out = np.zeros_like(a)
    ii = 0
    for i in range(len(a)-1,-1,-1):
        out[ii] = a[i]
        ii +=1
    return out


@njit
def normalize_matrix_columns(a):
    """Numba-jitted function to normalize the columns of a 2D array to one.

    Parameters
    ----------
    a: ~numpy.ndarray
        The 2D array to column-normalize.

    Returns
    -------
    ~numpy.ndarray
        The normalized array.
    """
    an = np.zeros_like(a)
    norm = np.zeros(a.shape[0])
    for i in range(a.shape[1]):
        norm[i] = np.linalg.norm(a[:, i], 2)
        an[:, i] = a[:, i] / norm[i]
    return an, norm


@njit
def solve_triangular_matrix(a, b):
    """Solve a triangular linear matrix equation :math:`ax = b`.
    
    Parameters
    ----------
    a: ~numpy.ndarray
        The 2D array :math:`a`.
    b: ~numpy.ndarray
        The 2D array :math:`b`.

    Returns
    -------
    ~numpy.ndarray
        The 2D array of solution :math:`x`.
    """
    x = np.zeros_like(a)
    for i in range(2, a.shape[0]+1):
        x[:i, i - 1] = np.linalg.solve(a[:i, :i], b[:i, i - 1])
    x[0, 0] = b[0, 0] / a[0, 0]
    return x



"""
    symbolic qgs tensor module
    ==========================

    This module computes and holds the symbolic representation of the tensors representing the tendencies of the model's
    equations.

"""
from qgs.functions.symbolic_mul import symbolic_tensordot
from qgs.functions.util import add_to_dict
from qgs.params.params import Parameter, ScalingParameter, ParametersArray, Params

import numpy as np
from sympy import simplify
import pickle

from sympy.matrices.immutable import ImmutableSparseMatrix
from sympy.tensor.array import ImmutableSparseNDimArray

# TODO: Check non stored IP version of this


class SymbolicQgsTensor(object):
    """Symbolic qgs tendencies tensor class.

    Parameters
    ----------
    params: None or QgParams, optional
        The models parameters to configure the tensor. `None` to initialize an empty tensor. Default to `None`.
    atmospheric_inner_products: None or AtmosphericInnerProducts, optional
        The inner products of the atmospheric basis functions on which the model's PDE atmospheric equations are
        projected.
        If `None`, disable the atmospheric tendencies. Default to `None`.
        The inner product is returned in symbolic or numeric form.
    oceanic_inner_products: None or OceanicInnerProducts, optional
        The inner products of the oceanic basis functions on which the model's PDE oceanic equations are projected.
        If `None`, disable the oceanic tendencies. Default to `None`.
        The inner product is returned in symbolic or numeric form.
    ground_inner_products: None or GroundInnerProducts, optional
        The inner products of the ground basis functions on which the model's PDE ground equations are projected.
        If `None`, disable the ground tendencies. Default to `None`.
        The inner product is returned in symbolic or numeric form.

    Attributes
    ----------
    params: None or QgParams
        The models parameters used to configure the tensor. `None` for an empty tensor.
    atmospheric_inner_products: None or AtmosphericInnerProducts
        The inner products of the atmospheric basis functions on which the model's PDE atmospheric equations are
        projected.
        If `None`, disable the atmospheric tendencies. Default to `None`.
    oceanic_inner_products: None or OceanicInnerProducts
        The inner products of the oceanic basis functions on which the model's PDE oceanic equations are projected.
        If `None`, disable the oceanic tendencies. Default to `None`.
    ground_inner_products: None or GroundInnerProducts
        The inner products of the ground basis functions on which the model's PDE ground equations are projected.
        If `None`, disable the ground tendencies. Default to `None`.
    tensor: ~sympy.tensor.array.ImmutableSparseNDimArray
        The tensor :math:`\\mathcal{T}_{i,j,k}` :math:`i`-th components.
    jacobian_tensor: ~sympy.tensor.array.ImmutableSparseNDimArray
        The jacobian tensor :math:`\\mathcal{T}_{i,j,k} + \\mathcal{T}_{i,k,j}` :math:`i`-th components.
    """

    def __init__(self, params=None, atmospheric_inner_products=None, oceanic_inner_products=None,
                 ground_inner_products=None):

        self.atmospheric_inner_products = atmospheric_inner_products
        self.oceanic_inner_products = oceanic_inner_products
        self.ground_inner_products = ground_inner_products
        self.params = params

        self.tensor = None
        self.jacobian_tensor = None

        if not self.params.dynamic_T:
            self.compute_tensor()

    def _psi_a(self, i):
        """Transform the :math:`\\psi_{\\mathrm a}` :math:`i`-th coefficient into the effective model's variable.

        Parameters
        ----------
        i: int
            The :math:`i`-th coefficients of :math:`\\psi_{\\mathrm a}`

        Returns
        -------
        int
            The effective model's variable.
        """

        return i + 1

    def _theta_a(self, i):
        """Transform the :math:`\\theta_{\\mathrm a}` :math:`i`-th coefficient into the effective model's variable.

        Parameters
        ----------
        i: int
            The :math:`i`-th coefficients of :math:`\\theta_{\\mathrm a}`

        Returns
        -------
        int
            The effective model's variable.
        """
        return i + self.params.variables_range[0] + 1

    def _psi_o(self, i):
        """Transform the :math:`\\psi_{\\mathrm o}` :math:`i`-th coefficient into the effective model's variable.

        Parameters
        ----------
        i: int
            The :math:`i`-th coefficients of :math:`\\psi_{\\mathrm o}`

        Returns
        -------
        int
            The effective model's variable.
        """
        return i + self.params.variables_range[1] + 1

    def _deltaT_o(self, i):
        """Transform the :math:`\\delta T_{\\mathrm o}` :math:`i`-th coefficient into the effective model's variable.

        Parameters
        ----------
        i: int
            The :math:`i`-th coefficients of :math:`\\delta T_{\\mathrm o}`

        Returns
        -------
        int
            The effective model's variable.
        """
        return i + self.params.variables_range[2] + 1

    def _deltaT_g(self, i):
        """Transform the :math:`\\delta T_{\\mathrm o}` :math:`i`-th coefficient into the effective model's variable.

        Parameters
        ----------
        i: int
            The :math:`i`-th coefficients of :math:`\\delta T_{\\mathrm o}`

        Returns
        -------
        int
            The effective model's variable.
        """
        return i + self.params.variables_range[1] + 1
    
    def _compute_tensor_dicts(self):
        if self.params is None:
            return None

        if self.atmospheric_inner_products is None and self.oceanic_inner_products is None \
                and self.ground_inner_products is None:
            return None

        aips = self.atmospheric_inner_products

        par = self.params
        atp = par.atemperature_params
        gp = par.ground_params
        ap = par.atmospheric_params

        nvar = par.number_of_variables

        bips = None
        if self.oceanic_inner_products is not None:
            bips = self.oceanic_inner_products
            ocean = True
        else:
            ocean = False

        if self.ground_inner_products is not None:
            bips = self.ground_inner_products
            ground_temp = True
        else:
            ground_temp = False

        if self.params.dynamic_T:
            offset = 1
        else:
            offset = 0

        # constructing some derived matrices
        if aips is not None:
            a_inv = dict()
            for i in range(offset, nvar[1]):
                for j in range(offset, nvar[1]):
                    a_inv[(i - offset, j - offset)] = aips.a(i, j)

            a_inv = ImmutableSparseMatrix(nvar[0], nvar[0], a_inv)
            a_inv = a_inv.inverse()
            a_inv = a_inv.simplify()

            a_theta = dict()
            for i in range(nvar[1]):
                for j in range(nvar[1]):
                    a_theta[(i, j)] = ap.sig0.symbol * aips.a(i, j) - aips.u(i, j)
            
            a_theta = ImmutableSparseMatrix(nvar[1], nvar[1], a_theta)
            a_theta = a_theta.inverse()
            a_theta = a_theta.simplify()

        if bips is not None:
            if ocean:
                U_inv = dict()
                for i in range(nvar[3]):
                    for j in range(nvar[3]):
                        U_inv[(i, j)] = bips.U(i, j)
                U_inv = ImmutableSparseMatrix(nvar[3], nvar[3], U_inv)
                U_inv = U_inv.inverse()
                U_inv = U_inv.simplify()

                M_psio = dict()
                for i in range(offset, nvar[3]):
                    for j in range(offset, nvar[3]):
                        M_psio[(i - offset, j - offset)] = bips.M(i, j) + self.params.G.symbolic_expression * bips.U(i, j)

                M_psio = ImmutableSparseMatrix(nvar[2], nvar[2], M_psio)
                M_psio = M_psio.inverse()
                M_psio = M_psio.simplify()

            else:
                U_inv = dict()
                for i in range(nvar[2]):
                    for j in range(nvar[2]):
                        U_inv[(i, j)] = bips.U(i, j)
                U_inv = ImmutableSparseMatrix(nvar[2], nvar[2], U_inv)
                U_inv = U_inv.inverse()
                U_inv = U_inv.simplify()

        ################

        if bips is not None:
            go = bips.stored
        else:
            go = True

        sy_arr_dic = dict()

        if aips.stored and go:
            # psi_a part
            a_inv_mult_c = a_inv @ aips._c[offset:, offset:]

            if gp is not None:
                hk_sym_arr = ImmutableSparseMatrix(gp.hk.symbols)

                if gp.orographic_basis == "atmospheric":
                    a_inv_mult_g = symbolic_tensordot(a_inv, aips._g[offset:, offset:, offset:], axes=1)
                    oro = symbolic_tensordot(a_inv_mult_g, hk_sym_arr, axes=1)[:, :, 0]
                else:
                    a_inv_mult_gh = symbolic_tensordot(a_inv, aips._gh[offset:, offset:, offset:], axes=1)
                    oro = symbolic_tensordot(a_inv_mult_gh, hk_sym_arr, axes=1)[:, :, 0]

            a_inv_mult_b = symbolic_tensordot(a_inv, aips._b[offset:, offset:, offset:], axes=1)
            
            if ocean:
                a_inv_mult_d = a_inv @ aips._d[offset:, offset:]

            for i in range(nvar[0]):
                for j in range(nvar[0]):
                    jo = j + offset  # skipping the theta 0 variable if it exists
                    sy_arr_dic = add_to_dict(sy_arr_dic, (self._psi_a(i), self._psi_a(j), 0), -a_inv_mult_c[i, j] * par.scale_params.beta.symbol)
                    sy_arr_dic = add_to_dict(sy_arr_dic, (self._psi_a(i), self._psi_a(j), 0), -(ap.kd.symbol * _kronecker_delta(i, j)) / 2)
                    sy_arr_dic = add_to_dict(sy_arr_dic, (self._psi_a(i), self._theta_a(jo), 0), (ap.kd.symbol * _kronecker_delta(i, j)) / 2)

                    if gp is not None:
                        # convert 
                        if gp.hk is not None:
                            sy_arr_dic = add_to_dict(sy_arr_dic, (self._psi_a(i), self._psi_a(j), 0), -oro[i, j] / 2)
                            sy_arr_dic = add_to_dict(sy_arr_dic, (self._psi_a(i), self._theta_a(jo), 0), oro[i, j] / 2)

                    for k in range(nvar[0]):
                        ko = k + offset  # skipping the theta 0 variable if it exists
                        sy_arr_dic = add_to_dict(sy_arr_dic, (self._psi_a(i), self._psi_a(j), self._psi_a(k)), -a_inv_mult_b[i, j, k])
                        sy_arr_dic = add_to_dict(sy_arr_dic, (self._psi_a(i), self._theta_a(jo), self._theta_a(ko)), -a_inv_mult_b[i, j, k])

                if ocean:
                    for j in range(nvar[2]):
                        sy_arr_dic = add_to_dict(sy_arr_dic, (self._psi_a(i), self._psi_o(j), 0), a_inv_mult_d[i, j] * ap.kd.symbol / 2)
            
            # theta_a part
            a_theta_mult_u = a_theta @ aips._u
            if self.params.Cpa is not None:
                val_Cpa = a_theta_mult_u @ ImmutableSparseMatrix(self.params.Cpa.symbolic_expressions)

            if atp.hd is not None and atp.thetas is not None:
                thetas_sym_arr = ImmutableSparseMatrix(atp.thetas.symbols)
                val_thetas = a_theta_mult_u @ thetas_sym_arr

            a_theta_mult_a = a_theta @ aips._a[:, offset:]
            a_theta_mult_c = a_theta @ aips._c[:, offset:]

            a_theta_mult_g = symbolic_tensordot(a_theta, aips._g[:, offset:, offset:], axes=1)

            if gp is not None:
                if gp.orographic_basis == "atmospheric":
                    oro = symbolic_tensordot(a_theta_mult_g, hk_sym_arr, axes=1)[:, :, 0]
                else:
                    a_theta_mult_gh = symbolic_tensordot(a_theta, aips._gh[:, offset:, offset:], axes=1)
                    oro = symbolic_tensordot(a_theta_mult_gh, hk_sym_arr, axes=1)[:, :, 0]

            a_theta_mult_b = symbolic_tensordot(a_theta, aips._b[:, offset:, offset:], axes=1)

            if ocean:
                a_theta_mult_d = a_theta @ aips._d[:, offset:]
                a_theta_mult_s = a_theta @ aips._s

            if ground_temp:
                a_theta_mult_s = a_theta @ aips._s
                
            for i in range(nvar[1]):
                if self.params.Cpa is not None:
                    sy_arr_dic = add_to_dict(sy_arr_dic, (self._theta_a(i), 0, 0), -val_Cpa[i])

                if atp.hd is not None and atp.thetas is not None:
                    sy_arr_dic = add_to_dict(sy_arr_dic, (self._theta_a(i), 0, 0), -val_thetas[i] * atp.hd.symbol)

                for j in range(nvar[0]):

                    jo = j + offset  # skipping the theta 0 variable if it exists

                    val_2 = a_theta_mult_a[i, j] * ap.kd.symbol * ap.sig0.symbol / 2
                    val_3 = a_theta_mult_a[i, j] * (ap.kd.symbol / 2 + 2 * ap.kdp.symbol) * ap.sig0.symbol
                    sy_arr_dic = add_to_dict(sy_arr_dic, (self._theta_a(i), self._psi_a(j), 0), val_2)
                    sy_arr_dic = add_to_dict(sy_arr_dic, (self._theta_a(i), self._theta_a(jo), 0), -val_3)

                    sy_arr_dic = add_to_dict(sy_arr_dic, (self._theta_a(i), self._theta_a(jo), 0), -a_theta_mult_c[i, j] * par.scale_params.beta.symbol * ap.sig0.symbol)

                    if gp is not None:
                        if gp.hk is not None:
                            sy_arr_dic = add_to_dict(sy_arr_dic, (self._theta_a(i), self._theta_a(jo), 0), -ap.sig0.symbol * oro[i, j] / 2)
                            sy_arr_dic = add_to_dict(sy_arr_dic, (self._theta_a(i), self._psi_a(j), 0), ap.sig0.symbol * oro[i, j] / 2)

                    for k in range(nvar[0]):
                        ko = k + offset  # skipping the theta 0 variable if it exists
                        sy_arr_dic = add_to_dict(sy_arr_dic, (self._theta_a(i), self._psi_a(j), self._theta_a(ko)), - a_theta_mult_b[i, j, k] * ap.sig0.symbol)
                        sy_arr_dic = add_to_dict(sy_arr_dic, (self._theta_a(i), self._theta_a(jo), self._psi_a(k)), - a_theta_mult_b[i, j, k] * ap.sig0.symbol)

                        sy_arr_dic = add_to_dict(sy_arr_dic, (self._theta_a(i), self._psi_a(j), self._theta_a(ko)), a_theta_mult_g[i, j, k])

                for j in range(nvar[1]):
                    if self.params.Lpa is not None:
                        sy_arr_dic = add_to_dict(sy_arr_dic, (self._theta_a(i), self._theta_a(j), 0), a_theta_mult_u[i, j] * atp.sc.symbol * self.params.Lpa.symbolic_expression)
                    if self.params.LSBpa is not None:
                        sy_arr_dic = add_to_dict(sy_arr_dic, (self._theta_a(i), self._theta_a(j), 0), a_theta_mult_u[i, j] * self.params.LSBpa.symbolic_expression)

                    if atp.hd is not None:
                        sy_arr_dic = add_to_dict(sy_arr_dic, (self._theta_a(i), self._theta_a(j), 0), a_theta_mult_u[i, j] * atp.hd)

                if ocean:
                    for j in range(nvar[2]):
                        sy_arr_dic = add_to_dict(sy_arr_dic, (self._theta_a(i), self._psi_o(j), 0), -a_theta_mult_d[i, j] * ap.sig0.symbol * ap.kd.symbol / 2)

                    if self.params.Lpa is not None:
                        for j in range(nvar[3]):
                            sy_arr_dic = add_to_dict(sy_arr_dic, (self._theta_a(i), self._deltaT_o(j), 0), -a_theta_mult_s[i, j] * self.params.Lpa.symbolic_expression / 2)
                            if self.params.LSBpgo is not None:
                                sy_arr_dic = add_to_dict(sy_arr_dic, (self._theta_a(i), self._deltaT_o(j), 0), -a_theta_mult_s[i, j] * self.params.LSBpgo.symbolic_expression)

                if ground_temp:
                    if self.params.Lpa is not None:
                        for j in range(nvar[2]):
                            sy_arr_dic = add_to_dict(sy_arr_dic, (self._theta_a(i), self._deltaT_g(j), 0), -a_theta_mult_s[i, j] * self.params.Lpa.symbolic_expression / 2)
                            if self.params.LSBpgo is not None:
                                sy_arr_dic = add_to_dict(sy_arr_dic, (self._theta_a(i), self._deltaT_g(j), 0), -a_theta_mult_s[i, j] * self.params.LSBpgo.symbolic_expression)

            if ocean:
                # psi_o part
                M_psio_mult_K = M_psio @ bips._K[offset:, offset:]
                M_psio_mult_N = M_psio @ bips._N[offset:, offset:]
                M_psio_mult_M = M_psio @ bips._M[offset:, offset:]
                M_psio_mult_C = symbolic_tensordot(M_psio, bips._C[offset:, offset:, offset:], axes=1)

                for i in range(nvar[2]):
                    for j in range(nvar[0]):
                        jo = j + offset  # skipping the theta 0 variable if it exists

                        sy_arr_dic = add_to_dict(sy_arr_dic, (self._psi_o(i), self._psi_a(j), 0), M_psio_mult_K[i, j] * par.oceanic_params.d.symbol)
                        sy_arr_dic = add_to_dict(sy_arr_dic, (self._psi_o(i), self._theta_a(jo), 0), -M_psio_mult_K[i, j] * par.oceanic_params.d.symbol)

                    for j in range(nvar[2]):
                        sy_arr_dic = add_to_dict(sy_arr_dic, (self._psi_o(i), self._psi_o(j), 0), -M_psio_mult_N[i, j] * par.scale_params.beta.symbol)

                        sy_arr_dic = add_to_dict(sy_arr_dic, (self._psi_o(i), self._psi_o(j), 0), -M_psio_mult_M[i, j] * (par.oceanic_params.r.symbol + par.oceanic_params.d.symbol))

                        for k in range(nvar[2]):
                            sy_arr_dic = add_to_dict(sy_arr_dic, (self._psi_o(i), self._psi_o(j), self._psi_o(k)), - M_psio_mult_C[i, j, k])

                # deltaT_o part
                U_inv_mult_W = U_inv @ bips._W
                Cpgo_sym_arr = ImmutableSparseMatrix(self.params.Cpgo.symbolic_expressions)
                U_inv_mult_W_Cpgo = U_inv_mult_W @ Cpgo_sym_arr

                U_inv_mult_O = symbolic_tensordot(U_inv, bips._O[:, offset:, offset:], axes=1)
                
                for i in range(nvar[3]):
                    sy_arr_dic = add_to_dict(sy_arr_dic, (self._deltaT_o(i), 0, 0), U_inv_mult_W_Cpgo[i])

                    for j in range(nvar[1]):
                        sy_arr_dic = add_to_dict(sy_arr_dic, (self._deltaT_o(i), self._theta_a(j), 0), U_inv_mult_W[i, j] * 2 * atp.sc.symbol * self.params.Lpgo.symbolic_expression)
                        if self.params.sbpa is not None:
                            sy_arr_dic = add_to_dict(sy_arr_dic, (self._deltaT_o(i), self._theta_a(j), 0), U_inv_mult_W[i, j] * self.params.sbpa.symbolic_expression)

                    for j in range(nvar[3]):
                        sy_arr_dic = add_to_dict(sy_arr_dic, (self._deltaT_o(i), self._deltaT_o(j), 0), - self.params.Lpgo.symbolic_expression * _kronecker_delta(i, j))
                        if self.params.sbpgo is not None:
                            sy_arr_dic = add_to_dict(sy_arr_dic, (self._deltaT_o(i), self._deltaT_o(j), 0), - self.params.sbpgo.symbolic_expression * _kronecker_delta(i, j))

                    for j in range(nvar[2]):
                        for k in range(nvar[2]):
                            ko = k + offset  # skipping the T 0 variable if it exists
                            sy_arr_dic = add_to_dict(sy_arr_dic, (self._deltaT_o(i), self._psi_o(j), self._deltaT_o(ko)), -U_inv_mult_O[i, j, k])

            # deltaT_g part
            if ground_temp:
                U_inv_mult_W = U_inv @ bips._W
                Cpgo_sym_arr = ImmutableSparseMatrix(self.params.Cpgo.symbolic_expressions)
                U_inv_mult_W_Cpgo = U_inv_mult_W @ Cpgo_sym_arr
                for i in range(nvar[2]):
                    sy_arr_dic = add_to_dict(sy_arr_dic, (self._deltaT_g(i), 0, 0), U_inv_mult_W_Cpgo[i])

                    for j in range(nvar[1]):
                        sy_arr_dic = add_to_dict(sy_arr_dic, (self._deltaT_g(i), self._theta_a(j), 0), U_inv_mult_W[i, j] * 2 * atp.sc.symbol * self.params.Lpgo.symbolic_expression)
                        if self.params.sbpa is not None:
                            sy_arr_dic = add_to_dict(sy_arr_dic, (self._deltaT_g(i), self._theta_a(j), 0), U_inv_mult_W[i, j] * self.params.sbpa.symbolic_expression)

                    for j in range(nvar[2]):
                        sy_arr_dic = add_to_dict(sy_arr_dic, (self._deltaT_g(i), self._deltaT_g(j), 0), - self.params.Lpgo.symbolic_expression * _kronecker_delta(i, j))
                        if self.params.sbpgo is not None:
                            sy_arr_dic = add_to_dict(sy_arr_dic, (self._deltaT_g(i), self._deltaT_g(j), 0), - self.params.sbpgo.symbolic_expression * _kronecker_delta(i, j))

        else:
            # psi_a part
            for i in range(nvar[0]):
                for j in range(nvar[0]):

                    jo = j + offset

                    val = 0
                    for jj in range(nvar[0]):
                        val += a_inv[i, jj] * aips.c(offset + jj, jo)
                    sy_arr_dic = add_to_dict(sy_arr_dic, (self._psi_a(i), self._psi_a(j), 0), - val * par.scale_params.beta.symbol)
                    sy_arr_dic = add_to_dict(sy_arr_dic, (self._psi_a(i), self._psi_a(j), 0), - (ap.kd.symbol * _kronecker_delta(i, j)) / 2)
                    sy_arr_dic = add_to_dict(sy_arr_dic, (self._psi_a(i), self._theta_a(jo), 0), (ap.kd.symbol * _kronecker_delta(i, j)) / 2)

                    if gp is not None:
                        hk_sym_arr = ImmutableSparseNDimArray(gp.hk.symbols)
                        if gp.hk is not None:
                            oro = 0
                            if gp.orographic_basis == "atmospheric":
                                for jj in range(nvar[0]):
                                    for kk in range(nvar[0]):
                                        oro += a_inv[i, jj] * aips.g(offset + jj, j, offset + kk) * hk_sym_arr[kk]
                            else:
                                for jj in range(nvar[0]):
                                    for kk in range(nvar[0]):
                                        oro += a_inv[i, jj] * aips.gh(offset + jj, j, offset + kk) * hk_sym_arr[kk]
                            sy_arr_dic = add_to_dict(sy_arr_dic, (self._psi_a(i), self._psi_a(j), 0), - oro / 2)
                            sy_arr_dic = add_to_dict(sy_arr_dic, (self._psi_a(i), self._theta_a(jo), 0), oro / 2)

                    for k in range(nvar[0]):
                        ko = k + offset  # skipping the theta 0 variable if it exists
                        val = 0
                        for jj in range(nvar[0]):
                            val += a_inv[i, jj] * aips.b(offset + jj, jo, ko)
                        sy_arr_dic = add_to_dict(sy_arr_dic, (self._psi_a(i), self._psi_a(j), self._psi_a(k)), - val)
                        sy_arr_dic = add_to_dict(sy_arr_dic, (self._psi_a(i), self._theta_a(jo), self._theta_a(ko)), - val)
                if ocean:
                    for j in range(nvar[2]):
                        jo = j + offset
                        val = 0
                        for jj in range(nvar[0]):
                            val += a_inv[i, jj] * aips.d(offset + jj, jo)
                        sy_arr_dic = add_to_dict(sy_arr_dic, (self._psi_a(i), self._psi_o(j), 0), val * ap.kd.symbol / 2)

            # theta_a part
            for i in range(nvar[1]):
                if self.params.Cpa is not None:
                    val = 0
                    for jj in range(nvar[1]):
                        for kk in range(nvar[1]):
                            val -= a_theta[i, jj] * aips.u(jj, kk) * self.params.Cpa.symbolic_expressions[kk]

                    sy_arr_dic = add_to_dict(sy_arr_dic, (self._theta_a(i), 0, 0), val)

                if atp.hd is not None and atp.thetas is not None:
                    thetas_sym_arr = ImmutableSparseNDimArray(atp.thetas.symbols)
                    val = 0
                    for jj in range(nvar[1]):
                        for kk in range(nvar[1]):
                            val -= a_theta[i, jj] * aips.u(jj, kk) * thetas_sym_arr[kk]

                    sy_arr_dic = add_to_dict(sy_arr_dic, (self._theta_a(i), 0, 0), val * atp.hd.symbol)

                for j in range(nvar[0]):

                    jo = j + offset  # skipping the theta 0 variable if it exists

                    val = 0
                    for jj in range(nvar[1]):
                        val += a_theta[i, jj] * aips.a(jj, jo)

                    sy_arr_dic = add_to_dict(sy_arr_dic, (self._theta_a(i), self._psi_a(j), 0), val * ap.kd.symbol * ap.sig0.symbol / 2)
                    sy_arr_dic = add_to_dict(sy_arr_dic, (self._theta_a(i), self._theta_a(jo), 0), - val * (ap.kd.symbol / 2 - 2 * ap.kdp.symbol) * ap.sig0.symbol)

                    val = 0
                    for jj in range(nvar[1]):
                        val -= a_theta[i, jj] * aips.c(jj, jo)

                    sy_arr_dic = add_to_dict(sy_arr_dic, (self._theta_a(i), self._theta_a(jo), 0), val * par.scale_params.beta.symbol * ap.sig0.symbol)

                    if gp is not None:
                        if gp.hk is not None:
                            oro = 0
                            if gp.orographic_basis == "atmospheric":
                                for jj in range(nvar[1]):
                                    for kk in range(nvar[0]):
                                        oro += a_theta[i, jj] * aips.g(jj, jo, offset + kk) * gp.hk[kk].symbol
                            else:
                                for jj in range(nvar[1]):
                                    for kk in range(nvar[0]):
                                        oro += a_theta[i, jj] * aips.gh(jj, jo, offset + kk) * gp.hk[kk].symbol
                            sy_arr_dic = add_to_dict(sy_arr_dic, (self._theta_a(i), self._theta_a(jo), 0), - ap.sig0.symbol * oro / 2)
                            sy_arr_dic = add_to_dict(sy_arr_dic, (self._theta_a(i), self._psi_a(j), 0), ap.sig0.symbol * oro / 2)

                    for k in range(nvar[0]):
                        ko = k + offset  # skipping the theta 0 variable if it exists
                        val = 0
                        for jj in range(nvar[1]):
                            val += a_theta[i, jj] * aips.b(jj, jo, ko)

                        sy_arr_dic = add_to_dict(sy_arr_dic, (self._theta_a(i), self._psi_a(j), self._theta_a(ko)), - val * ap.sig0.symbol)
                        sy_arr_dic = add_to_dict(sy_arr_dic, (self._theta_a(i), self._theta_a(jo), self._psi_a(k)), - val * ap.sig0.symbol)

                        val = 0
                        for jj in range(nvar[1]):
                            val += a_theta[i, jj] * aips.g(jj, jo, ko)

                        sy_arr_dic = add_to_dict(sy_arr_dic, (self._theta_a(i), self._psi_a(j), self._theta_a(ko)), val)

                for j in range(nvar[1]):
                    val = 0
                    for jj in range(nvar[1]):
                        val += a_theta[i, jj] * aips.u(jj, j)

                    if self.params.Lpa is not None:
                        sy_arr_dic = add_to_dict(sy_arr_dic, (self._theta_a(i), self._theta_a(j), 0), val * atp.sc.symbol * self.params.Lpa.symbolic_expression)
                    
                    if self.params.LSBpa is not None:
                        sy_arr_dic = add_to_dict(sy_arr_dic, (self._theta_a(i), self._theta_a(j), 0), val * self.params.LSBpa.symbolic_expression)

                    if atp.hd is not None:
                        sy_arr_dic = add_to_dict(sy_arr_dic, (self._theta_a(i), self._theta_a(j), 0), val * atp.hd.symbol)

                if ocean:
                    for j in range(nvar[2]):
                        jo = j + offset  # skipping the theta 0 variable if it exists
                        val = 0
                        for jj in range(nvar[1]):
                            val -= a_theta[i, jj] * aips.d(jj, jo)
                        
                        sy_arr_dic = add_to_dict(sy_arr_dic, (self._theta_a(i), self._psi_o(j), 0), val * ap.sig0.symbol * ap.kd.symbol / 2)

                    if self.params.Lpa is not None:
                        for j in range(nvar[3]):
                            val = 0
                            for jj in range(nvar[1]):
                                val -= a_theta[i, jj] * aips.s(jj, j)
                            sy_arr_dic = add_to_dict(sy_arr_dic, (self._theta_a(i), self._deltaT_o(j), 0), val * self.params.Lpa.symbolic_expression / 2)
                            if self.params.LSBpgo is not None:
                                sy_arr_dic = add_to_dict(sy_arr_dic, (self._theta_a(i), self._deltaT_o(j), 0), val * self.params.LSBpgo.symbolic_expression)

                if ground_temp:
                    if self.params.Lpa is not None:
                        for j in range(nvar[2]):
                            val = 0
                            for jj in range(nvar[1]):
                                val -= a_theta[i, jj] * aips.s(jj, j)
                            sy_arr_dic = add_to_dict(sy_arr_dic, (self._theta_a(i), self._deltaT_g(j), 0), val * self.params.Lpa.symbolic_expression / 2)
                            if self.params.LSBpgo is not None:
                                sy_arr_dic = add_to_dict(sy_arr_dic, (self._theta_a(i), self._deltaT_g(j), 0), val * self.params.LSBpgo.symbolic_expression)

            if ocean:
                # psi_o part
                for i in range(nvar[2]):
                    for j in range(nvar[0]):
                        jo = j + offset  # skipping the theta 0 variable if it exists

                        for jj in range(nvar[2]):
                            val = M_psio[i, jj] * bips.K(offset + jj, jo) * par.oceanic_params.d.symbol
                        sy_arr_dic = add_to_dict(sy_arr_dic, (self._psi_o(i), self._psi_a(j), 0), val)
                        sy_arr_dic = add_to_dict(sy_arr_dic, (self._psi_o(i), self._theta_a(jo), 0), - val)

                    for j in range(nvar[2]):
                        jo = j + offset  # skipping the T 0 variable if it exists

                        val = 0
                        for jj in range(nvar[2]):
                            val -= M_psio[i, jj] * bips.N(offset + jj, jo)
                        sy_arr_dic = add_to_dict(sy_arr_dic, (self._psi_o(i), self._psi_o(j), 0), val * par.scale_params.beta.symbol)

                        val = 0
                        for jj in range(nvar[2]):
                            val -= M_psio[i, jj] * bips.M(offset + jj, jo)
                        sy_arr_dic = add_to_dict(sy_arr_dic, (self._psi_o(i), self._psi_o(j), 0), val * (par.oceanic_params.r.symbol + par.oceanic_params.d.symbol))

                        for k in range(nvar[2]):
                            ko = k + offset  # skipping the T 0 variable if it exists
                            val = 0
                            for jj in range(nvar[2]):
                                val -= M_psio[i, jj] * bips.C(offset + jj, jo, ko)
                            sy_arr_dic = add_to_dict(sy_arr_dic, (self._psi_o(i), self._psi_o(j), self._psi_o(k)), val)

                # deltaT_o part
                for i in range(nvar[3]):
                    val = 0
                    for jj in range(nvar[1]):
                        for kk in range(nvar[3]):
                            val += U_inv[i, kk] * bips.W(kk, jj) * self.params.Cpgo.symbolic_expressions[jj]
                    
                    sy_arr_dic = add_to_dict(sy_arr_dic, (self._deltaT_o(i), 0, 0), val)

                    for j in range(nvar[1]):
                        val = 0
                        for jj in range(nvar[3]):
                            val += U_inv[i, jj] * bips.W(jj, j)
                        sy_arr_dic = add_to_dict(sy_arr_dic, (self._deltaT_o(i), self._theta_a(j), 0), val * 2 * atp.sc.symbol * self.params.Lpgo.symbolic_expression)
                        if self.params.sbpa is not None:
                            sy_arr_dic = add_to_dict(sy_arr_dic, (self._deltaT_o(i), self._theta_a(j), 0), val * self.params.sbpa.symbolic_expression)

                    for j in range(nvar[3]):
                        sy_arr_dic = add_to_dict(sy_arr_dic, (self._deltaT_o(i), self._deltaT_o(j), 0), - self.params.Lpgo.symbolic_expression * _kronecker_delta(i, j))
                        if self.params.sbpgo is not None:
                            sy_arr_dic = add_to_dict(sy_arr_dic, (self._deltaT_o(i), self._deltaT_o(j), 0), - self.params.sbpgo.symbolic_expression * _kronecker_delta(i, j))

                    for j in range(nvar[2]):
                        for k in range(offset, nvar[3]):
                            jo = j + offset  # skipping the T 0 variable if it exists

                            val = 0
                            for jj in range(nvar[3]):
                                val -= U_inv[i, jj] * bips.O(jj, jo, k)
                            sy_arr_dic = add_to_dict(sy_arr_dic, (self._deltaT_o(i), self._psi_o(j), self._deltaT_o(k)), val)

            # deltaT_g part
            if ground_temp:
                for i in range(nvar[2]):
                    val = 0
                    for jj in range(nvar[1]):
                        for kk in range(nvar[2]):
                            val += U_inv[i, kk] * bips.W(kk, jj) * self.params.Cpgo.symbolic_expressions[jj]
                    
                    sy_arr_dic = add_to_dict(sy_arr_dic, (self._deltaT_g(i), 0, 0), val)

                    for j in range(nvar[1]):
                        val = 0
                        for jj in range(nvar[2]):
                            val += U_inv[i, jj] * bips.W(jj, j)
                        sy_arr_dic = add_to_dict(sy_arr_dic, (self._deltaT_g(i), self._theta_a(j), 0), val * 2 * atp.sc.symbol * self.params.Lpgo.symbolic_expression)
                        
                        if self.params.sbpa is not None:
                            sy_arr_dic = add_to_dict(sy_arr_dic, (self._deltaT_g(i), self._theta_a(j), 0), val * self.params.sbpa.symbolic_expression)

                    for j in range(nvar[2]):
                        sy_arr_dic = add_to_dict(sy_arr_dic, (self._deltaT_g(i), self._deltaT_g(j), 0), - self.params.Lpgo.symbolic_expression * _kronecker_delta(i, j))
                        if self.params.sbpgo is not None:
                            sy_arr_dic = add_to_dict(sy_arr_dic, (self._deltaT_g(i), self._deltaT_g(j), 0), - self.params.sbpgo.symbolic_expression * _kronecker_delta(i, j))

        return sy_arr_dic

    def compute_tensor(self):
        """Routine to compute the tensor."""
        sy_arr_dic = self._compute_tensor_dicts()
        sy_arr_dic = self.remove_dic_zeros(sy_arr_dic)

        if sy_arr_dic is not None:
            self._set_tensor(sy_arr_dic)

    def _set_tensor(self, dic, set_symbolic=False):
        self.jac_dic = self.remove_dic_zeros(self.jacobian_from_dict(dic))
        self.tensor_dic = self.remove_dic_zeros(self.simplify_dict(dic))
        
        if set_symbolic:
            self._set_symbolic_tensor()
            
    def _set_symbolic_tensor(self):
        ndim = self.params.ndim

        if self.params.dynamic_T:
            if self.params.T4:
                raise ValueError("Symbolic tensor output not configured for T4 version, use Dynamic T version")
            else:
                dims = (ndim + 1, ndim + 1, ndim + 1, ndim + 1, ndim + 1)
        else:
            dims = (ndim + 1, ndim + 1, ndim + 1)

        jacobian_tensor = ImmutableSparseNDimArray(self.jac_dic.copy(), dims)
        tensor = ImmutableSparseNDimArray(self.tensor_dic.copy(), dims)

        self.jacobian_tensor = jacobian_tensor.applyfunc(simplify)
        self.tensor = tensor.applyfunc(simplify)

    @staticmethod
    def remove_dic_zeros(dic):
        """Removes zero values from dictionary

        Parameters
        ----------
        tensor: dict
            dictionary which could include 0 in values
        Returns
        -------
        ten_out: dict
            dictionary with same keys and values as input, but keys with value of 0 are removed
        """

        non_zero_dic = dict()
        for key in dic.keys():
            if dic[key] != 0:
                non_zero_dic[key] = dic[key]

        return non_zero_dic
    
    @staticmethod
    def jacobian_from_dict(dic):
        """Calculates the Jacobian from the qgs tensor

        Parameters
        ----------
        dic: dict
            dictionary of tendencies of the model
        Returns
        -------
        dic_jac: dict
            Jacobian tensor stored in a dictionary
        """

        rank = max([len(i) for i in dic.keys()])
        n_perm = rank - 2
        
        orig_order = [i for i in range(rank)]
        
        keys = dic.keys()
        dic_jac = dic.copy()
        
        for i in range(1, n_perm+1):
            new_pos = orig_order.copy()
            new_pos[1] = orig_order[i+1]
            new_pos[i+1] = orig_order[1]
            for key in keys:
                dic_jac = add_to_dict(dic_jac, tuple(key[i] for i in new_pos), dic[key])
        
        return dic_jac
    
    @staticmethod
    def simplify_dict(dic):
        """calculates the upper triangular tensor of a given tensor, stored in dictionary

        Parameters
        ----------
        dic: dict
            dictionary of tendencies of the model

        Returns
        -------
        dic_upp: dict
            Upper triangular tensor, stored as a tensor where the keys are the coordinates of the corresponding value.
        """

        keys = dic.keys()
        dic_upp = dict()

        for key in keys:
            new_key = tuple([key[0]] + sorted(key[1:]))
            dic_upp = add_to_dict(dic_upp, new_key, dic[key])

        return dic_upp
    
    def save_to_file(self, filename, **kwargs):
        """Function to save the tensor object to a file with the :mod:`pickle` module.

        Parameters
        ----------
        filename: str
            The file name where to save the tensor object.
        kwargs: dict
            Keyword arguments to pass to the :mod:`pickle` module method.
        """
        f = open(filename, 'wb')
        pickle.dump(self.__dict__, f, **kwargs)
        f.close()

    def sub_tensor(self, tensor=None, continuation_variables=None):
        """Uses sympy substitution to convert the symbolic tensor or a symbolic dictionary to a numerical one.

        Parameters
        ----------
        tensor: dict(~sympy.core.expr.Expr or float) or ~sympy.tensor.array.ImmutableSparseNDimArray
            Tensor of model tendencies, either as

            - a dictionary with keys of non-zero coordinates, and values of Sympy expressions or floats
            - or a sparse Sympy tensor

        continuation_variables: list(Parameter, ScalingParameter or ParametersArray) or `None`
            Variables which remain symbolic, all other variables are substituted with numerical values.
            If `None` all variables are substituted.

        Returns
        -------
        ten_out: dict(float)
            Dictionary of the substituted tensor of the model tendencies, with coordinates and numerical values
        """

        if continuation_variables is None:
            continuation_variables = list()

        param_subs = _parameter_substitutions(self.params, continuation_variables)
        
        if tensor is None:
            ten = self.tensor_dic
        else:
            ten = tensor

        if isinstance(ten, dict):
            ten_out = dict()
            for key in ten.keys():
                val = ten[key].subs(param_subs)
                try:
                    ten_out[key] = float(val)
                except:
                    ten_out[key] = val

        else:
            # Assuming the tensor is a sympy tensor
            ten_out = ten.subs(param_subs)

        return ten_out
            
    def print_tensor(self, tensor=None, dict_opp=True, tol=1e-10):
        """Print the non-zero coordinates of values of the tensor of the model tendencies

        Parameters
        ----------
        tensor: dict(~sympy.core.expr.Expr or float) or ~sympy.tensor.array.ImmutableSparseNDimArray or `None`
            Tensor of model tendencies, either as

            - a dictionary with keys of non-zero coordinates, and values of Sympy expressions or floats
            - or a sparse Sympy tensor

            If `None`, defaults to the stored tensor.
            Defaults to `None`.

        dict_opp: bool
            If `True`, returns the unsimplified symbolic expressions, if `False` the simplified expressions are returned.
        tol: float
            The tolerance to allow for numerical errors when finding non-zero values.
        """

        if tensor is None:
            if dict_opp:
                temp_ten = self.tensor_dic
            else:
                temp_ten = self.tensor
        else:
            temp_ten = tensor

        if isinstance(temp_ten, dict):
            val_list = [(key, temp_ten[key]) for key in temp_ten.keys()] 
        else:
            val_list = np.ndenumerate(temp_ten)

        for ix, v in val_list:
            if isinstance(v, float):
                bool_test = (abs(v) > tol)
            else:
                bool_test = (v != 0)

            if bool_test:
                try:
                    output_val = float(v)
                except:
                    try:
                        output_val = v.simplify().evalf()
                    except:
                        output_val = v
                print(str(ix) + ": " + str(output_val))


class SymbolicQgsTensorDynamicT(SymbolicQgsTensor):
    """qgs dynamical temperature first order (linear) symbolic tendencies tensor class.

    Parameters
    ----------
    params: None or QgParams, optional
        The models parameters to configure the tensor. `None` to initialize an empty tensor. Default to `None`.
    atmospheric_inner_products: None or AtmosphericInnerProducts, optional
        The inner products of the atmospheric basis functions on which the model's PDE atmospheric equations are projected.
        If `None`, disable the atmospheric tendencies. Default to `None`.
    oceanic_inner_products: None or OceanicInnerProducts, optional
        The inner products of the oceanic basis functions on which the model's PDE oceanic equations are projected.
        If `None`, disable the oceanic tendencies. Default to `None`.
    ground_inner_products: None or GroundInnerProducts, optional
        The inner products of the ground basis functions on which the model's PDE ground equations are projected.
        If `None`, disable the ground tendencies. Default to `None`.

    Attributes
    ----------
    params: None or QgParams
        The models parameters used to configure the tensor. `None` for an empty tensor.
    atmospheric_inner_products: None or AtmosphericInnerProducts
        The inner products of the atmospheric basis functions on which the model's PDE atmospheric equations are projected.
        If `None`, disable the atmospheric tendencies. Default to `None`.
    oceanic_inner_products: None or OceanicInnerProducts
        The inner products of the oceanic basis functions on which the model's PDE oceanic equations are projected.
        If `None`, disable the oceanic tendencies. Default to `None`.
    ground_inner_products: None or GroundInnerProducts
        The inner products of the ground basis functions on which the model's PDE ground equations are projected.
        If `None`, disable the ground tendencies. Default to `None`.
    tensor: ~sympy.tensor.array.ImmutableSparseNDimArray
        The tensor :math:`\\mathcal{T}_{i,j,k}` :math:`i`-th components.
    jacobian_tensor: ~sympy.tensor.array.ImmutableSparseNDimArray
        The jacobian tensor :math:`\\mathcal{T}_{i,j,k} + \\mathcal{T}_{i,k,j}` :math:`i`-th components.
    """

    def __init__(self, params=None, atmospheric_inner_products=None, oceanic_inner_products=None, ground_inner_products=None):

        SymbolicQgsTensor.__init__(self, params, atmospheric_inner_products, oceanic_inner_products, ground_inner_products)
        
        if params.dynamic_T:
            self.compute_tensor()

    def _compute_tensor_dicts(self):

        if self.params is None:
            return None

        if self.atmospheric_inner_products is None and self.oceanic_inner_products is None \
                and self.ground_inner_products is None:
            return None

        aips = self.atmospheric_inner_products

        bips = None
        if self.oceanic_inner_products is not None:
            bips = self.oceanic_inner_products

        elif self.ground_inner_products is not None:
            bips = self.ground_inner_products

        if bips is not None:
            go = bips.stored
        else:
            go = True

        if aips.stored and go:
            symbolic_array_full_dict = self._compute_stored_full_dict()

        else:
            symbolic_array_full_dict = self._compute_non_stored_full_dict()

        return symbolic_array_full_dict

    def _compute_stored_full_dict(self):
        par = self.params
        nvar = par.number_of_variables
        aips = self.atmospheric_inner_products

        bips = None
        if self.oceanic_inner_products is not None:
            bips = self.oceanic_inner_products
            ocean = True
        else:
            ocean = False

        if self.ground_inner_products is not None:
            bips = self.ground_inner_products
            ground_temp = True
        else:
            ground_temp = False

        # constructing some derived matrices
        if aips is not None:

            a_theta = dict()
            for i in range(nvar[1]):
                for j in range(nvar[1]):
                    a_theta[(i, j)] = par.atmospheric_params.sig0.symbol * aips.a(i, j) - aips.u(i, j)
            a_theta = ImmutableSparseMatrix(nvar[1], nvar[1], a_theta)
            a_theta = a_theta.inverse()

        if bips is not None:
            U_inv = dict()
            if ocean:
                for i in range(nvar[3]):
                    for j in range(nvar[3]):
                        U_inv[(i, j)] = bips.U(i, j)
                U_inv = ImmutableSparseMatrix(nvar[3], nvar[3], U_inv)
                U_inv = U_inv.inverse()
            else:
                for i in range(nvar[2]):
                    for j in range(nvar[2]):
                        U_inv[i, j] = bips.U(i, j)
                U_inv = ImmutableSparseMatrix(nvar[2], nvar[2], U_inv)
                U_inv = U_inv.inverse()

        #################

        sy_arr_dic = dict()
        # theta_a part
        for i in range(nvar[1]):
            if self.params.T4LSBpa is not None:
                j = k = ell = 0
                for m in range(nvar[1]):
                    val = 0
                    for jj in range(nvar[1]):
                        val += a_theta[i, jj] * aips._z[jj, j, k, ell, m]
                    if m == 0:
                        sy_arr_dic = add_to_dict(sy_arr_dic, (self._theta_a(i), self._theta_a(j), self._theta_a(k), self._theta_a(ell), self._theta_a(m)), self.params.T4LSBpa.symbolic_expression * val)
                    else:
                        sy_arr_dic = add_to_dict(sy_arr_dic, (self._theta_a(i), self._theta_a(j), self._theta_a(k), self._theta_a(ell), self._theta_a(m)), 4 * self.params.T4LSBpa.symbolic_expression * val)
                        
            if ocean:
                if self.params.T4LSBpgo is not None:
                    j = k = ell = 0
                    for m in range(nvar[3]):
                        val = 0
                        for jj in range(nvar[1]):
                            val += a_theta[i, jj] * aips._v[jj, j, k, ell, m]
                        if m == 0:
                            sy_arr_dic = add_to_dict(sy_arr_dic, (self._theta_a(i), self._deltaT_o(j), self._deltaT_o(k), self._deltaT_o(ell), self._deltaT_o(m)), -self.params.T4LSBpgo.symbolic_expression * val)
                        else:
                            sy_arr_dic = add_to_dict(sy_arr_dic, (self._theta_a(i), self._deltaT_o(j), self._deltaT_o(k), self._deltaT_o(ell), self._deltaT_o(m)), -4 * self.params.T4LSBpgo.symbolic_expression * val)

            if ground_temp:
                if self.params.T4LSBpgo is not None:
                    j = k = ell = 0
                    for m in range(nvar[2]):
                        val = 0
                        for jj in range(nvar[1]):
                            val += a_theta[i, jj] * aips._v[jj, j, k, ell, m]
                        if m == 0:
                            sy_arr_dic = add_to_dict(sy_arr_dic, (self._theta_a(i), self._deltaT_g(j), self._deltaT_g(k), self._deltaT_g(ell), self._deltaT_g(m)), -self.params.T4LSBpgo.symbolic_expression * val)
                        else:
                            sy_arr_dic = add_to_dict(sy_arr_dic, (self._theta_a(i), self._deltaT_g(j), self._deltaT_g(k), self._deltaT_g(ell), self._deltaT_g(m)), -4 * self.params.T4LSBpgo.symbolic_expression * val)

        if ocean:
            # delta_T part
            for i in range(nvar[3]):
                j = k = ell = 0
                for m in range(nvar[1]):
                    val = 0
                    for jj in range(nvar[3]):
                        val += U_inv[i, jj] * bips._Z[jj, j, k, ell, m]
                    if m == 0:
                        sy_arr_dic = add_to_dict(sy_arr_dic, (self._deltaT_o(i), self._theta_a(j), self._theta_a(k), self._theta_a(ell), self._theta_a(m)), self.params.T4sbpa.symbolic_expression * val)
                    else:
                        sy_arr_dic = add_to_dict(sy_arr_dic, (self._deltaT_o(i), self._theta_a(j), self._theta_a(k), self._theta_a(ell), self._theta_a(m)), 4 * self.params.T4sbpa.symbolic_expression * val)
                    
                for m in range(nvar[3]):
                    val = 0
                    for jj in range(nvar[3]):
                        val += U_inv[i, jj] * bips._V[jj, j, k, ell, m]
                    if m == 0:
                        sy_arr_dic = add_to_dict(sy_arr_dic, (self._deltaT_o(i), self._deltaT_o(j), self._deltaT_o(k), self._deltaT_o(ell), self._deltaT_o(m)), - self.params.T4sbpgo.symbolic_expression * val)
                    else:
                        sy_arr_dic = add_to_dict(sy_arr_dic, (self._deltaT_o(i), self._deltaT_o(j), self._deltaT_o(k), self._deltaT_o(ell), self._deltaT_o(m)), -4 * self.params.T4sbpgo.symbolic_expression * val)

        if ground_temp:
            # deltaT_g part
            for i in range(nvar[2]):
                j = k = ell = 0
                for m in range(nvar[1]):
                    val = 0
                    for jj in range(nvar[2]):
                        val += U_inv[i, jj] * bips._Z[jj, j, k, ell, m]
                    if m == 0:
                        sy_arr_dic = add_to_dict(sy_arr_dic, (self._deltaT_g(i), self._theta_a(j), self._theta_a(k), self._theta_a(ell), self._theta_a(m)), self.params.T4sbpa.symbolic_expression * val)
                    else:
                        sy_arr_dic = add_to_dict(sy_arr_dic, (self._deltaT_g(i), self._theta_a(j), self._theta_a(k), self._theta_a(ell), self._theta_a(m)), 4 * self.params.T4sbpa.symbolic_expression * val)

                for m in range(nvar[2]):
                    val = 0 
                    for jj in range(nvar[2]):
                        val += U_inv[i, jj] * bips._V[jj, j, k, ell, m]
                    if m == 0:
                        sy_arr_dic = add_to_dict(sy_arr_dic, (self._deltaT_g(i), self._deltaT_g(j), self._deltaT_g(k), self._deltaT_g(ell), self._deltaT_g(m)), -self.params.T4sbpgo.symbolic_expression * val)
                    else:
                        sy_arr_dic = add_to_dict(sy_arr_dic, (self._deltaT_g(i), self._deltaT_g(j), self._deltaT_g(k), self._deltaT_g(ell), self._deltaT_g(m)), -4 * self.params.T4sbpgo.symbolic_expression * val)
                    
        return sy_arr_dic

    def _compute_non_stored_full_dict(self):
        par = self.params
        nvar = par.number_of_variables
        aips = self.atmospheric_inner_products

        bips = None
        if self.oceanic_inner_products is not None:
            bips = self.oceanic_inner_products
            ocean = True
        else:
            ocean = False

        if self.ground_inner_products is not None:
            bips = self.ground_inner_products
            ground_temp = True
        else:
            ground_temp = False

        # constructing some derived matrices
        if aips is not None:
            a_theta = dict()
            for i in range(nvar[1]):
                for j in range(nvar[1]):
                    a_theta[(i, j)] = par.atmospheric_params.sig0.symbol * aips.a(i, j) - aips.u(i, j)

            a_theta = ImmutableSparseMatrix(nvar[1], nvar[1], a_theta)
            a_theta = a_theta.inverse()

        if bips is not None:
            if ocean:
                U_inv = dict()
                for i in range(nvar[3]):
                    for j in range(nvar[3]):
                        U_inv[i, j] = bips.U(i, j)
                U_inv = ImmutableSparseMatrix(nvar[3], nvar[3], U_inv)

            else:
                U_inv = dict()
                for i in range(nvar[2]):
                    for j in range(nvar[2]):
                        U_inv[(i, j)] = bips.U(i, j)
                U_inv = ImmutableSparseMatrix(nvar[2], nvar[2], U_inv)
            
            U_inv = U_inv.inverse()
                
        #################

        sy_arr_dic = dict()
        # theta_a part
        for i in range(nvar[1]):

            if self.params.T4LSBpa is not None:
                j = k = ell = 0
                for m in range(nvar[1]):
                    val = 0
                    for jj in range(nvar[1]):
                        val += a_theta[i, jj] * aips.z(jj, j, k, ell, m)
                    if m == 0:
                        sy_arr_dic[(self._theta_a(i), self._theta_a(j), self._theta_a(k), self._theta_a(ell), self._theta_a(m))] = self.params.T4LSBpa.symbolic_expression * val
                    else:
                        sy_arr_dic[(self._theta_a(i), self._theta_a(j), self._theta_a(k), self._theta_a(ell), self._theta_a(m))] = 4 * self.params.T4LSBpa.symbolic_expression * val

            if ocean:
                if self.params.T4LSBpgo is not None:
                    j = k = ell = 0
                    for m in range(nvar[3]):
                        val = 0
                        for jj in range(nvar[1]):
                            val -= a_theta[i, jj] * aips.v(jj, j, k, ell, m)
                        if m == 0:
                            sy_arr_dic[(self._theta_a(i), self._deltaT_o(j), self._deltaT_o(k), self._deltaT_o(ell), self._deltaT_o(m))] = self.params.T4LSBpgo.symbolic_expression * val
                        else:
                            sy_arr_dic[(self._theta_a(i), self._deltaT_o(j), self._deltaT_o(k), self._deltaT_o(ell), self._deltaT_o(m))] = 4 * self.params.T4LSBpgo.symbolic_expression * val

            if ground_temp:
                if self.params.T4LSBpgo is not None:
                    j = k = ell = 0
                    for m in range(nvar[2]):
                        val = 0
                        for jj in range(nvar[1]):
                            val -= a_theta[i, jj] * aips.v(jj, j, k, ell, m)
                        if m == 0:
                            sy_arr_dic[(self._theta_a(i), self._deltaT_g(j), self._deltaT_g(k), self._deltaT_g(ell), self._deltaT_g(m))] = self.params.T4LSBpgo.symbolic_expression * val
                        else:
                            sy_arr_dic[(self._theta_a(i), self._deltaT_g(j), self._deltaT_g(k), self._deltaT_g(ell), self._deltaT_g(m))] = 4 * self.params.T4LSBpgo.symbolic_expression * val

        if ocean:

            # deltaT_o part
            for i in range(nvar[3]):

                j = k = ell = 0
                for m in range(nvar[1]):
                    val = 0
                    for jj in range(nvar[3]):
                        val += U_inv[i, jj] * bips.Z(jj, j, k, ell, m)
                    if m == 0:
                        sy_arr_dic[(self._deltaT_o(i), self._theta_a(j), self._theta_a(k), self._theta_a(ell), self._theta_a(m))] = self.params.T4sbpa.symbolic_expression * val
                    else:
                        sy_arr_dic[(self._deltaT_o(i), self._theta_a(j), self._theta_a(k), self._theta_a(ell), self._theta_a(m))] = 4 * self.params.T4sbpa.symbolic_expression * val

                for m in range(nvar[3]):
                    val = 0
                    for jj in range(nvar[3]):
                        val -= U_inv[i, jj] * bips.V(jj, j, k, ell, m)
                    if m == 0:
                        sy_arr_dic[(self._deltaT_o(i), self._deltaT_o(j), self._deltaT_o(k), self._deltaT_o(ell), self._deltaT_o(m))] = self.params.T4sbpgo.symbolic_expression * val
                    else:
                        sy_arr_dic[(self._deltaT_o(i), self._deltaT_o(j), self._deltaT_o(k), self._deltaT_o(ell), self._deltaT_o(m))] = 4 * self.params.T4sbpgo.symbolic_expression * val

        # deltaT_g part
        if ground_temp:
            for i in range(nvar[2]):

                j = k = ell = 0
                for m in range(nvar[1]):
                    val = 0
                    for jj in range(nvar[2]):
                        val += U_inv[i, jj] * bips._Z[jj, j, k, ell, m]
                    if m == 0:
                        sy_arr_dic[(self._deltaT_g(i), self._theta_a(j), self._theta_a(k), self._theta_a(ell), self._theta_a(m))] = self.params.T4sbpa.symbolic_expression * val
                    else:
                        sy_arr_dic[(self._deltaT_g(i), self._theta_a(j), self._theta_a(k), self._theta_a(ell), self._theta_a(m))] = 4 * self.params.T4sbpa.symbolic_expression * val

                for m in range(nvar[2]):
                    val = 0
                    for jj in range(nvar[2]):
                        val -= U_inv[i, jj] * bips._V[jj, j, k, ell, m]
                    if m == 0:
                        sy_arr_dic[(self._deltaT_g(i), self._deltaT_g(j), self._deltaT_g(k), self._deltaT_g(ell), self._deltaT_g(m))] = self.params.T4sbpgo.symbolic_expression * val
                    else:
                        sy_arr_dic[(self._deltaT_g(i), self._deltaT_g(j), self._deltaT_g(k), self._deltaT_g(ell), self._deltaT_g(m))] = 4 * self.params.T4sbpgo.symbolic_expression * val

        return sy_arr_dic

    def compute_tensor(self):
        """Routine to compute the tensor."""
        if self.params.T4:
            # TODO: Make a proper error message for here
            raise ValueError("Parameters are set for T4 version, set dynamic_T=True")

        symbolic_dict_linear = SymbolicQgsTensor._compute_tensor_dicts(self)
        symbolic_dict_linear = _shift_dict_keys(symbolic_dict_linear, (0, 0))

        symbolic_dict_dynT = self._compute_tensor_dicts()

        if symbolic_dict_linear is not None:
            symbolic_dict_dynT = {**symbolic_dict_linear, **symbolic_dict_dynT}

        if symbolic_dict_dynT is not None:
            self._set_tensor(symbolic_dict_dynT)


class SymbolicQgsTensorT4(SymbolicQgsTensor):
    """qgs dynamical temperature first order (linear) symbolic tendencies tensor class.

    Parameters
    ----------
    params: None or QgParams, optional
        The models parameters to configure the tensor. `None` to initialize an empty tensor. Default to `None`.
    atmospheric_inner_products: None or AtmosphericInnerProducts, optional
        The inner products of the atmospheric basis functions on which the model's PDE atmospheric equations are projected.
        If `None`, disable the atmospheric tendencies. Default to `None`.
    oceanic_inner_products: None or OceanicInnerProducts, optional
        The inner products of the oceanic basis functions on which the model's PDE oceanic equations are projected.
        If `None`, disable the oceanic tendencies. Default to `None`.
    ground_inner_products: None or GroundInnerProducts, optional
        The inner products of the ground basis functions on which the model's PDE ground equations are projected.
        If `None`, disable the ground tendencies. Default to `None`.

    Attributes
    ----------
    params: None or QgParams
        The models parameters used to configure the tensor. `None` for an empty tensor.
    atmospheric_inner_products: None or AtmosphericInnerProducts
        The inner products of the atmospheric basis functions on which the model's PDE atmospheric equations are projected.
        If `None`, disable the atmospheric tendencies. Default to `None`.
    oceanic_inner_products: None or OceanicInnerProducts
        The inner products of the oceanic basis functions on which the model's PDE oceanic equations are projected.
        If `None`, disable the oceanic tendencies. Default to `None`.
    ground_inner_products: None or GroundInnerProducts
        The inner products of the ground basis functions on which the model's PDE ground equations are projected.
        If `None`, disable the ground tendencies. Default to `None`.
    tensor: ~sympy.tensor.array.ImmutableSparseNDimArray
        The tensor :math:`\\mathcal{T}_{i,j,k}` :math:`i`-th components.
    jacobian_tensor: ~sympy.tensor.array.ImmutableSparseNDimArray
        The jacobian tensor :math:`\\mathcal{T}_{i,j,k} + \\mathcal{T}_{i,k,j}` :math:`i`-th components.
    """

    def __init__(self, params=None, atmospheric_inner_products=None, oceanic_inner_products=None, ground_inner_products=None):

        SymbolicQgsTensor.__init__(self, params, atmospheric_inner_products, oceanic_inner_products, ground_inner_products)
        
        if params.T4:
            self.compute_tensor()

    def _compute_non_stored_full_dict(self):
        par = self.params
        nvar = par.number_of_variables
        aips = self.atmospheric_inner_products

        bips = None
        if self.oceanic_inner_products is not None:
            bips = self.oceanic_inner_products
            ocean = True
        else:
            ocean = False

        if self.ground_inner_products is not None:
            bips = self.ground_inner_products
            ground_temp = True
        else:
            ground_temp = False

        # constructing some derived matrices
        if aips is not None:
            a_theta = dict()
            for i in range(nvar[1]):
                for j in range(nvar[1]):
                    a_theta[(i, j)] = par.atmospheric_params.sig0.symbol * aips.a(i, j) - aips.u(i, j)

            a_theta = ImmutableSparseMatrix(nvar[1], nvar[1], a_theta)
            a_theta = a_theta.inverse()

        if bips is not None:
            if ocean:
                U_inv = dict()
                for i in range(nvar[3]):
                    for j in range(nvar[3]):
                        U_inv[i, j] = bips.U(i, j)
                U_inv = ImmutableSparseMatrix(nvar[3], nvar[3], U_inv)

            else:
                U_inv = dict()
                for i in range(nvar[2]):
                    for j in range(nvar[2]):
                        U_inv[(i, j)] = bips.U(i, j)
                U_inv = ImmutableSparseMatrix(nvar[2], nvar[2], U_inv)
            
            U_inv = U_inv.inverse()
                
        #################

        sy_arr_dic = dict()
        # theta_a part
        for i in range(nvar[1]):

            if self.params.T4LSBpa is not None:
                for j in range(nvar[1]):
                    for k in range(nvar[1]):
                        for ell in range(nvar[1]):
                            for m in range(nvar[1]):
                                val = 0
                                for jj in range(nvar[1]):
                                    val += a_theta[i, jj] * aips.z(jj, j, k, ell, m)
                                
                                sy_arr_dic[(self._theta_a(i), self._theta_a(j), self._theta_a(k), self._theta_a(ell), self._theta_a(m))] = self.params.T4LSBpa.symbolic_expression * val
                    
            if ocean:
                if self.params.T4LSBpgo is not None:
                    for j in range(nvar[3]):
                        for k in range(nvar[3]):
                            for ell in range(nvar[3]):
                                for m in range(nvar[3]):
                                    val = 0
                                    for jj in range(nvar[1]):
                                        val -= a_theta[i, jj] * aips.v(jj, j, k, ell, m)
                                    
                                    sy_arr_dic[(self._theta_a(i), self._deltaT_o(j), self._deltaT_o(k), self._deltaT_o(ell), self._deltaT_o(m))] = self.params.T4LSBpgo.symbolic_expression * val

            if ground_temp:
                if self.params.T4LSBpgo is not None:
                    for j in range(nvar[2]):
                        for k in range(nvar[2]):
                            for ell in range(nvar[2]):
                                for m in range(nvar[2]):
                                    val = 0
                                    for jj in range(nvar[1]):
                                        val -= a_theta[i, jj] * aips.v(jj, j, k, ell, m)

                                    sy_arr_dic[(self._theta_a(i), self._deltaT_g(j), self._deltaT_g(k), self._deltaT_g(ell), self._deltaT_g(m))] = self.params.T4LSBpgo.symbolic_expression * val
                        
        if ocean:

            # deltaT_o part
            for i in range(nvar[3]):
                for j in range(nvar[1]):
                    for k in range(nvar[1]):
                        for ell in range(nvar[1]):
                            for m in range(nvar[1]):
                                val = 0
                                for jj in range(nvar[3]):
                                    val += U_inv[i, jj] * bips.Z(jj, j, k, ell, m)
                                
                                sy_arr_dic[(self._deltaT_o(i), self._theta_a(j), self._theta_a(k), self._theta_a(ell), self._theta_a(m))] = self.params.T4LSBpa.symbolic_expression * val
                
                for j in range(nvar[3]):
                    for k in range(nvar[3]):
                        for ell in range(nvar[3]):
                            for m in range(nvar[3]):
                                val = 0
                                for jj in range(nvar[3]):
                                    val -= U_inv[i, jj] * bips.V(jj, j, k, ell, m)

                                sy_arr_dic[(self._deltaT_o(i), self._deltaT_o(j), self._deltaT_o(k), self._deltaT_o(ell), self._deltaT_o(m))] = self.params.T4LSBpgo.symbolic_expression * val

        # deltaT_g part
        if ground_temp:
            for i in range(nvar[2]):
                for j in range(nvar[1]):
                    for k in range(nvar[1]):
                        for ell in range(nvar[1]):
                            for m in range(nvar[1]):
                                val = 0
                                for jj in range(nvar[2]):
                                    val += U_inv[i, jj] * bips._Z[jj, j, k, ell, m]
                                
                                sy_arr_dic[(self._deltaT_g(i), self._theta_a(j), self._theta_a(k), self._theta_a(ell), self._theta_a(m))] = self.params.T4LSBpa.symbolic_expression * val
                
                for j in range(nvar[2]):
                    for k in range(nvar[2]):
                        for ell in range(nvar[2]):    
                            for m in range(nvar[2]):
                                val = 0
                                for jj in range(nvar[2]):
                                    val -= U_inv[i, jj] * bips._V[jj, j, k, ell, m]

                                sy_arr_dic[(self._deltaT_g(i), self._deltaT_g(j), self._deltaT_g(k), self._deltaT_g(ell), self._deltaT_g(m))] = self.params.T4LSBpgo.symbolic_expression * val
                   
        return sy_arr_dic

    def compute_tensor(self):
        """Routine to compute the tensor."""
        if not self.params.T4:
            raise ValueError("Parameters are not set for T4 version")

        symbolic_dict_linear = SymbolicQgsTensor._compute_tensor_dicts(self)
        symbolic_dict_linear = _shift_dict_keys(symbolic_dict_linear, (0, 0))

        symbolic_dict_T4 = self._compute_non_stored_full_dict()

        if symbolic_dict_linear is not None:
            symbolic_dict_T4 = {**symbolic_dict_linear, **symbolic_dict_T4}
        
        if symbolic_dict_T4 is not None:
            self._set_tensor(symbolic_dict_T4)


def _kronecker_delta(i, j):
    if i == j:
        return 1
    else:
        return 0


def _shift_dict_keys(dic, shift):
    """
    Keys of given dictionary are altered to add values in the given indicies

    Parameters
    ----------
    dic: dict
        Dictionary representing a tensor.

    shift: tuple
        A tuple that represents the shift in the tensor indices.
    """

    shifted_dic = dict()
    for key in dic.keys():
        new_key = key + shift
        shifted_dic[new_key] = dic[key]
    
    return shifted_dic


def _parameter_substitutions(params, continuation_variables):
    """
    Returns a dict of parameters values that are to be substituted,
    removing the parameters given in `continuation_variables`.
    """

    subs = params._all_items

    if continuation_variables is None:
        continuation_variables = list()

    # Remove variables in continuation variables
    for cv in continuation_variables:
        if isinstance(cv, ParametersArray):
            for cv_i in cv:
                subs.remove(cv_i)
        elif isinstance(cv, Parameter):
            subs.remove(cv)
        else:  # Try ... who knows...
            subs.remove(cv)

    # make the remaining items into a dict to pass to sympy subs function
    sub_dic = {}
    for p in subs:
        if p.symbol is not None:
            sub_dic[p.symbol] = float(p)
    return sub_dic


if __name__ == "__main__":
    dic = dict()
    dic = add_to_dict(dic, (0, 0), 1)
    dic = add_to_dict(dic, (0, 0), 2)
    print(dic)

    from qgs.params.params import QgParams
    from qgs.inner_products import symbolic

    params = QgParams({'rr': 287.e0, 'sb': 5.6e-8})
    params.set_atmospheric_channel_fourier_modes(6, 6, mode="symbolic")
    params.atmospheric_params.set_params({'sigma': 0.2, 'kd': 0.1, 'kdp': 0.01})

    params.ground_params.set_orography(0.2, 1)
    params.atemperature_params.set_thetas(0.1, 0)

    aip = symbolic.AtmosphericSymbolicInnerProducts(params, return_symbolic=True, make_substitution=True)

    # sym_aotensor = SymbolicQgsTensor(params=params, atmospheric_inner_products=aip)




"""
    Inner products definition module
    ================================

    Module containing classes to define the `inner products`_ used by the model.

    .. _inner products: https://en.wikipedia.org/wiki/Inner_product_space
    
"""

from abc import ABC, abstractmethod

# from sympy.simplify import trigsimp
from sympy.simplify.fu import TR8, TR10
from sympy import Symbol, diff, integrate, symbols, pi, Integral

_n = Symbol('n', positive=True)
_x, _y = symbols('x y')


class InnerProductDefinition(ABC):
    """Base class to define the model's basis inner products.

    Parameters
    ----------
    optimizer: None or callable, optional
        A function to optimize the computation of the integrals or the integrand.
        If `None`, does not optimize.

    Attributes
    ----------
    optimizer: None or callable
        A function to optimize the computation of the integrals or the integrand.
        If `None`, does not optimize the computation.
    """

    def __init__(self, optimizer=None):

        self.optimizer = None

        if optimizer is not None:
            self.set_optimizer(optimizer)
        else:
            self.set_optimizer(self._no_optimizer)

    def set_optimizer(self, optimizer):
        """Function to set the optimizer.

        Parameters
        ----------
        optimizer: callable
            A function to optimize the computation of the integrals or the integrand.
        """
        self.optimizer = optimizer

    @staticmethod
    def _no_optimizer(expr):
        return expr

    @staticmethod
    @abstractmethod
    def jacobian(S, G):
        """Jacobian present in the advection terms:

        Parameters
        ----------
        S:
            1st argument the Jacobian.
        G:
            2nd argument the Jacobian.
        """
        pass

    @staticmethod
    @abstractmethod
    def laplacian(S):
        """Laplacian :math:`\\nabla^2 S` of a function :math:`S`."""
        pass

    @abstractmethod
    def ip_lap(self, S, G, symbolic_expr=False):
        """Function to compute the inner product :math:`(S, \\nabla^2 G)`.

        Parameters
        ----------
        S:
            Left-hand side function of the product.
        G:
            Right-hand side function of the product.
        symbolic_expr: bool, optional
            If `True`, return the integral as a symbolic expression object. Else, return the integral performed symbolically.
        """
        pass

    @abstractmethod
    def ip_diff_x(self, S, G, symbolic_expr=False):
        """Function to compute the inner product :math:`(S, \\partial_x G)`.

        Parameters
        ----------
        S:
            Left-hand side function of the product.
        G:
            Right-hand side function of the product.
        symbolic_expr: bool, optional
            If `True`, return the integral as a symbolic expression object. Else, return the integral performed symbolically.
        """
        pass

    @abstractmethod
    def ip_jac(self, S, G, H, symbolic_expr=False):
        """Function to compute the inner product :math:`(S, J(G, H))`, where
        :math:`J` is the :meth:`jacobian`.

        Parameters
        ----------
        S:
            Left-hand side function of the product.
        G:
            1st argument of the right-hand side function of the product.
        H:
            2nd argument of the right-hand side function of the product.
        symbolic_expr: bool, optional
            If `True`, return the integral as a symbolic expression object. Else, return the integral performed symbolically.
        """
        pass

    @abstractmethod
    def ip_jac_lap(self, S, G, H, symbolic_expr=False):
        """Function to compute the inner product :math:`(S, J(G, \\nabla^2 H))`, where
        :math:`J` is the :meth:`jacobian`.

        Parameters
        ----------
        S:
            Left-hand side function of the product.
        G:
            1st argument of the right-hand side function of the product.
        H:
            2nd argument of the right-hand side function of the product.
        symbolic_expr: bool, optional
            If `True`, return the integral as a symbolic expression object. Else, return the integral performed symbolically.
        """
        pass


class SymbolicInnerProductDefinition(InnerProductDefinition):
    """Base class to define symbolic inner products using `Sympy`_.

    Parameters
    ----------
    optimizer: None or callable, optional
        A function to simplify the integrand to optimize the computation of the integrals.
        Should return a `Sympy`_ expression.
        If `None`, does not optimize the computation.

    Attributes
    ----------
    optimizer: None or callable
        A function to simplify the integrand to optimize the computation of the integrals.
        Should return a `Sympy`_ expression.
        If `None`, does not optimize the computation.


    .. _Sympy: https://www.sympy.org/

    """

    def __init__(self, optimizer=None):

        InnerProductDefinition.__init__(self, optimizer)

    @abstractmethod
    def symbolic_inner_product(self, S, G, symbolic_expr=False, integrand=False):
        """Symbolic definition of the inner product :math:`(S, G)`.

        Parameters
        ----------
        S: Sympy expression
            Left-hand side function of the product.
        G: Sympy expression
            Right-hand side function of the product.
        symbolic_expr: bool, optional
            If `True`, return the integral as a symbolic expression object. Else, return the integral performed symbolically.
        integrand: bool, optional
            If `True`, return the integrand of the integral and its integration limits as a list of symbolic expression object. Else, return the integral performed symbolically.

        Returns
        -------
        Sympy expression
            The symbolic result of the inner product.
        """
        pass

    def ip_lap(self, S, G, symbolic_expr=False, integrand=False):
        """Function to compute the inner product :math:`(S, \\nabla^2 G)`.

        Parameters
        ----------
        S: Sympy expression
            Left-hand side function of the product.
        G: Sympy expression
            Right-hand side function of the product.
        symbolic_expr: bool, optional
            If `True`, return the integral as a symbolic expression object. Else, return the integral performed symbolically.
        integrand: bool, optional
            If `True`, return the integrand of the integral and its integration limits as a list of symbolic expression object. Else, return the integral performed symbolically.

        Returns
        -------
        Sympy expression
            The symbolic result of the inner product.
        """
        return self.symbolic_inner_product(S, self.laplacian(G), symbolic_expr=symbolic_expr, integrand=integrand)

    def ip_diff_x(self, S, G, symbolic_expr=False, integrand=False):
        """Function to compute the inner product :math:`(S, \\partial_x G)`.

        Parameters
        ----------
        S: Sympy expression
            Left-hand side function of the product.
        G: Sympy expression
            Right-hand side function of the product.
        symbolic_expr: bool, optional
            If `True`, return the integral as a symbolic expression object. Else, return the integral performed symbolically.
        integrand: bool, optional
            If `True`, return the integrand of the integral and its integration limits as a list of symbolic expression object. Else, return the integral performed symbolically.

        Returns
        -------
        Sympy expression
            The symbolic result of the inner product.
        """
        return self.symbolic_inner_product(S, diff(G, _x), symbolic_expr=symbolic_expr, integrand=integrand)

    def ip_jac(self, S, G, H, symbolic_expr=False, integrand=False):
        """Function to compute the inner product :math:`(S, J(G, H))`, where
        :math:`J` is the :meth:`jacobian`.

        Parameters
        ----------
        S:  Sympy expression
            Left-hand side function of the product.
        G:  Sympy expression
            1st argument of the right-hand side function of the product.
        H:  Sympy expression
            2nd argument of the right-hand side function of the product.
        symbolic_expr: bool, optional
            If `True`, return the integral as a symbolic expression object. Else, return the integral performed symbolically.
        integrand: bool, optional
            If `True`, return the integrand of the integral and its integration limits as a list of symbolic expression object. Else, return the integral performed symbolically.

        Returns
        -------
        Sympy expression
            The symbolic result of the inner product.
        """
        return self.symbolic_inner_product(S, self.jacobian(G, H), symbolic_expr=symbolic_expr, integrand=integrand)

    def ip_jac_lap(self, S, G, H, symbolic_expr=False, integrand=False):
        """Function to compute the inner product :math:`(S, J(G, \\nabla^2 H))`, where
        :math:`J` is the :meth:`jacobian`.

        Parameters
        ----------
        S:  Sympy expression
            Left-hand side function of the product.
        G:  Sympy expression
            1st argument of the right-hand side function of the product.
        H:  Sympy expression
            2nd argument of the right-hand side function of the product.
        symbolic_expr: bool, optional
            If `True`, return the integral as a symbolic expression object. Else, return the integral performed symbolically.
        integrand: bool, optional
            If `True`, return the integrand of the integral and its integration limits as a list of symbolic expression object. Else, return the integral performed symbolically.

        Returns
        -------
        Sympy expression
            The symbolic result of the inner product.
        """
        return self.symbolic_inner_product(S, self.jacobian(G, self.laplacian(H)), symbolic_expr=symbolic_expr, integrand=integrand)


class StandardSymbolicInnerProductDefinition(SymbolicInnerProductDefinition):
    """Standard qgs class to define symbolic inner products using `Sympy`_.

    Parameters
    ----------
    optimizer: None or callable, optional
        A function to simplify the integrand to optimize the computation of the integrals.
        Should return a `Sympy`_ expression.
        If `None`, does not optimize the computation.

    Attributes
    ----------
    optimizer: None or callable
        A function to simplify the integrand to optimize the computation of the integrals.
        Should return a `Sympy`_ expression.
        If `None`, does not optimize the computation.


    .. _Sympy: https://www.sympy.org/

    """

    def __init__(self, optimizer=None):

        if optimizer is None:
            SymbolicInnerProductDefinition.__init__(self, self._trig_optimizer)
        else:
            SymbolicInnerProductDefinition.__init__(self, optimizer)

    @staticmethod
    def _trig_optimizer(expr):
        return TR10(TR8(expr))

    @staticmethod
    def jacobian(S, G):
        """Jacobian present in the advection terms:

        .. math:

            J(S, G) = \\partial_x S\\, \\partial_y G - \\partial_y S\\, \\partial_x G

        Parameters
        ----------
        S: Sympy expression
            1st argument the Jacobian.
        G: Sympy expression
            2nd argument the Jacobian.

        Returns
        -------
        Sympy expression
            The Jacobian.
        """
        return diff(S, _x) * diff(G, _y) - diff(G, _x) * diff(S, _y)

    @staticmethod
    def laplacian(S):
        """Laplacian :math:`\\nabla^2 S` of a given function :math:`S` in 2D cartesian coordinates:
        :math:`\\nabla^2 S = \\partial^2_x S + \\partial^2_y S`.

        Parameters
        ----------
        S: Sympy expression
            Functions to take the Laplacian of.

        Returns
        -------
        Sympy expression
            The Laplacian.
        """
        return diff(S, _x, 2) + diff(S, _y, 2)

    @staticmethod
    def integrate_over_domain(expr, symbolic_expr=False):
        """Definition of the normalized integrals over the spatial domain used by the inner products:
        :math:`\\frac{n}{2\\pi^2}\\int_0^\\pi\\int_0^{2\\pi/n} \\, \\mathrm{expr}(x, y) \\, \\mathrm{d} x \\, \\mathrm{d} y`.

        Parameters
        ----------
        expr: Sympy expression
            The expression to integrate.
        symbolic_expr: bool, optional
            If `True`, return the integral as a symbolic expression object. Else, return the integral performed symbolically.

        Returns
        -------
        Sympy expression
            The result of the symbolic integration.
        """
        if symbolic_expr:
            return Integral(expr, (_x, 0, 2 * pi / _n), (_y, 0, pi))
        else:
            return integrate(expr, (_x, 0, 2 * pi / _n), (_y, 0, pi))

    def symbolic_inner_product(self, S, G, symbolic_expr=False, integrand=False):
        """Function defining the inner product to be computed symbolically:
        :math:`(S, G) = \\frac{n}{2\\pi^2}\\int_0^\\pi\\int_0^{2\\pi/n} S(x,y)\\, G(x,y)\\, \\mathrm{d} x \\, \\mathrm{d} y`.

        Parameters
        ----------
        S: Sympy expression
            Left-hand side function of the product.
        G: Sympy expression
            Right-hand side function of the product.
        symbolic_expr: bool, optional
            If `True`, return the integral as a symbolic expression object. Else, return the integral performed symbolically.
        integrand: bool, optional
            If `True`, return the integrand of the integral and its integration limits as a list of symbolic expression object. Else, return the integral performed symbolically.

        Returns
        -------
        Sympy expression
            The result of the symbolic integration
        """
        expr = (_n / (2 * pi ** 2)) * S * G
        if integrand:
            return expr, (_x, 0, 2 * pi / _n), (_y, 0, pi)
        else:
            return self.integrate_over_domain(self.optimizer(expr), symbolic_expr=symbolic_expr)
